#!/usr/bin/env python
# The Typer parameters get mixed up if we use the __future__ import annotations in the main file.
# ruff: noqa: E402
import re
import sys
import traceback
from datetime import datetime, timezone
from pathlib import Path
from typing import NoReturn

import typer
from cognite.client.config import global_config
from rich.markup import escape
from rich.panel import Panel

from cognite_toolkit._cdf_tk.hints import Hint

# Do not warn the user about feature previews from the Cognite-SDK we use in Toolkit
global_config.disable_pypi_version_check = True
global_config.silence_feature_preview_warnings = True

from rich import print

from cognite_toolkit._cdf_tk.apps import (
    AuthApp,
    CoreApp,
    DumpApp,
    LandingApp,
    MigrateApp,
    ModulesApp,
    PopulateApp,
    ProfileApp,
    PurgeApp,
    RepoApp,
    RunApp,
)
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import (
    CollectCommand,
)
from cognite_toolkit._cdf_tk.constants import HINT_LEAD_TEXT, URL, USE_SENTRY
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitError,
)
from cognite_toolkit._cdf_tk.feature_flags import FeatureFlag, Flags
from cognite_toolkit._cdf_tk.plugins import Plugins
from cognite_toolkit._cdf_tk.tracker import Tracker
from cognite_toolkit._cdf_tk.utils import (
    sentry_exception_filter,
)
from cognite_toolkit._version import __version__ as current_version

if USE_SENTRY:
    import sentry_sdk

    sentry_sdk.init(
        dsn="https://20552f92b525fe551e9adc939024d526@o4508040730968064.ingest.de.sentry.io/4508160801374288",
        release=current_version,
        before_send=sentry_exception_filter,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        traces_sample_rate=1.0,
    )


CDF_TOML = CDFToml.load(Path.cwd())

default_typer_kws = dict(
    pretty_exceptions_short=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)
try:
    typer.Typer(**default_typer_kws)  # type: ignore [arg-type]
except AttributeError as e:
    # From Typer version 0.11 -> 0.12, breaks if you have an existing installation.
    raise ToolkitError(
        "'cognite-toolkit' uses a dependency named 'typer'. From 'typer' version 0.11 -> 0.12 there was a "
        "breaking change if you have an existing installation of 'typer'. The workaround is to uninstall "
        "'typer-slim', and then, reinstall 'typer':\n"
        "pip uninstall typer-slim\n"
        "pip install typer\n\n"
        f"This was triggered by the error: {e!r}"
    )

_app = CoreApp(**default_typer_kws)

user_app = typer.Typer(**default_typer_kws, hidden=True)  # type: ignore [arg-type]
landing_app = LandingApp(**default_typer_kws)  # type: ignore [arg-type]

_app.add_typer(AuthApp(**default_typer_kws), name="auth")
if Plugins.run.value.is_enabled():
    _app.add_typer(RunApp(**default_typer_kws), name="run")
_app.add_typer(RepoApp(**default_typer_kws), name="repo")

if Plugins.dump.value.is_enabled():
    _app.add_typer(DumpApp(**default_typer_kws), name="dump")

if Plugins.purge.value.is_enabled():
    _app.add_typer(PurgeApp(**default_typer_kws), name="purge")

if Flags.POPULATE.is_enabled():
    _app.add_typer(PopulateApp(**default_typer_kws), name="populate")

if Flags.PROFILE.is_enabled():
    _app.add_typer(ProfileApp(**default_typer_kws), name="profile")

if Flags.MIGRATE.is_enabled():
    _app.add_typer(MigrateApp(**default_typer_kws), name="migrate")

_app.add_typer(ModulesApp(**default_typer_kws), name="modules")
_app.command("init")(landing_app.main_init)


def app() -> NoReturn:
    # --- Main entry point ---
    # Users run 'app()' directly, but that doesn't allow us to control excepton handling:
    try:
        if FeatureFlag.is_enabled(Flags.IMPORT_CMD):
            from cognite_toolkit._cdf_tk.prototypes.import_app import import_app

            _app.add_typer(import_app, name="import")

        _app()
    except ToolkitError as err:
        if "--verbose" in sys.argv:
            print(Panel(traceback.format_exc(), title="Traceback", expand=False))

        print(f"  [bold red]ERROR ([/][red]{type(err).__name__}[/][bold red]):[/] {err}")
        raise SystemExit(1)
    except SystemExit:
        if result := re.search(r"click.exceptions.UsageError: No such command '(\w+)'.", traceback.format_exc()):
            cmd = result.group(1)
            if cmd in Plugins.list():
                plugin = r"[plugins]"
                print(
                    f"{HINT_LEAD_TEXT} The plugin [bold]{cmd}[/bold] is not enabled."
                    f"\nEnable it in the [bold]cdf.toml[/bold] file by setting '{cmd} = true' in the "
                    f"[bold]{escape(plugin)}[/bold] section."
                    f"\nDocs to learn more: {Hint.link(URL.plugins, URL.plugins)}"
                )
        raise

    raise SystemExit(0)


@_app.command("collect", hidden=True)
def collect(
    action: str = typer.Argument(
        help="Whether to explicitly opt-in or opt-out of usage data collection. [opt-in, opt-out]"
    ),
) -> None:
    """Collect usage information for the toolkit."""
    cmd = CollectCommand()
    cmd.run(lambda: cmd.execute(action))  # type: ignore [arg-type]


@user_app.callback(invoke_without_command=True)
def user_main(ctx: typer.Context) -> None:
    """Commands to give information about the toolkit."""
    if ctx.invoked_subcommand is None:
        print("Use [bold yellow]cdf user --help[/] to see available commands.")
    return None


@user_app.command("info")
def user_info() -> None:
    """Print information about user"""
    tracker = Tracker()
    print(f"ID={tracker.get_distinct_id()!r}\nnow={datetime.now(timezone.utc).isoformat(timespec='seconds')!r}")


if __name__ == "__main__":
    app()



from __future__ import annotations

import itertools
import os
import re
from abc import ABC
from collections import UserDict, defaultdict
from collections.abc import Hashable, Iterable, Sequence, Set
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, Literal, cast, get_args

import yaml
from rich import print

from cognite_toolkit._cdf_tk.constants import (
    _RUNNING_IN_BROWSER,
    BUILD_ENVIRONMENT_FILE,
    BUILTIN_MODULES,
    DEFAULT_CONFIG_FILE,
    DEFAULT_ENV,
    MODULES,
    ROOT_MODULES,
    SEARCH_VARIABLES_SUFFIX,
    EnvType,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitEnvError, ToolkitMissingModuleError
from cognite_toolkit._cdf_tk.hints import ModuleDefinition
from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME, RawDatabaseLoader
from cognite_toolkit._cdf_tk.tk_warnings import (
    FileReadWarning,
    MediumSeverityWarning,
    MissingFileWarning,
    SourceFileModifiedWarning,
    ToolkitWarning,
    WarningList,
)
from cognite_toolkit._cdf_tk.utils import (
    YAMLComment,
    YAMLWithComments,
    calculate_hash,
    flatten_dict,
    read_yaml_content,
    safe_read,
)
from cognite_toolkit._cdf_tk.utils.modules import parse_user_selected_modules
from cognite_toolkit._version import __version__

from . import BuiltModuleList
from ._base import ConfigCore, _load_version_variable
from ._built_resources import BuiltResourceList
from ._module_directories import ModuleDirectories, ReadModule

_AVAILABLE_ENV_TYPES = tuple(get_args(EnvType))


@dataclass
class Environment:
    name: str = "dev"
    project: str = field(default_factory=lambda: os.environ.get("CDF_PROJECT", "UNKNOWN"))
    validation_type: str = "dev"
    selected: list[str | Path] = field(default_factory=lambda: [Path(MODULES)])

    @property
    def is_strict_validation(self) -> bool:
        return self.validation_type.casefold() != "dev"

    @classmethod
    def load(cls, data: dict[str, Any], build_name: str) -> Environment:
        _deprecation_selected(data)
        if "name" not in data:
            data["name"] = build_name
        _deprecate_type(data, build_name)
        if missing := {"name", "project", "validation-type", "selected"} - set(data.keys()):
            raise ToolkitEnvError(
                f"Environment section is missing one or more required fields: {missing} in {BuildConfigYAML.get_filename(build_name)!s}"
            )
        return Environment(
            name=build_name,
            project=data["project"],
            validation_type=data["validation-type"],
            selected=parse_user_selected_modules(data.get("selected")),
        )

    def dump(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "project": self.project,
            "validation-type": self.validation_type,
            "selected": [
                selected.as_posix() + "/" if isinstance(selected, Path) else selected for selected in self.selected
            ],
        }

    def get_selected_modules(self, modules_by_package: dict[str, list[str]]) -> set[str | Path]:
        selected_modules: set[str | Path] = set()
        for selected in self.selected:
            if selected in modules_by_package and isinstance(selected, str):
                selected_modules.update(modules_by_package[selected])
            else:
                selected_modules.add(selected)
        return selected_modules


@dataclass
class ConfigYAMLCore(ABC):
    environment: Environment = field(default_factory=Environment)


@dataclass
class BuildConfigYAML(ConfigYAMLCore, ConfigCore):
    """This is the config.[env].yaml file used in the cdf-tk build command."""

    filename: ClassVar[str] = "config.{build_env}.yaml"
    variables: dict[str, Any] = field(default_factory=dict)

    def set_environment_variables(self) -> None:
        os.environ["CDF_ENVIRON"] = self.environment.name
        os.environ["CDF_BUILD_TYPE"] = self.environment.validation_type

    def validate_environment(self) -> ToolkitWarning | None:
        if _RUNNING_IN_BROWSER:
            return None
        project = self.environment.project
        project_env = os.environ.get("CDF_PROJECT")
        if project_env == project:
            return None

        is_strict_validation = self.environment.is_strict_validation
        env_name = self.environment.name
        file_name = self.get_filename(env_name)
        missing_message = (
            "No 'CDF_PROJECT' environment variable set. This is expected to match the project "
            f"set in environment section of {file_name!r}.\nThis is required for "
            "building configurations for staging and prod environments to ensure that you do "
            "not accidentally deploy to the wrong project."
        )
        mismatch_message = (
            f"Project name mismatch between project set in the environment section of {file_name!r} and the "
            f"environment variable 'CDF_PROJECT', {project} ≠ {project_env}.\nThis is required for "
            "building configurations for staging and prod environments to ensure that you do not "
            "accidentally deploy to the wrong project."
        )
        if is_strict_validation and project_env is None:
            raise ToolkitEnvError(missing_message)
        elif is_strict_validation:
            raise ToolkitEnvError(mismatch_message)
        elif not is_strict_validation and project_env is None:
            return MediumSeverityWarning(missing_message)
        else:
            return MediumSeverityWarning(mismatch_message)

    @classmethod
    def load(cls, data: dict[str, Any], build_env_name: str, filepath: Path) -> BuildConfigYAML:
        if "environment" not in data:
            err_msg = f"Expected 'environment' section in {filepath!s}."
            raise ToolkitEnvError(err_msg)
        environment = Environment.load(data["environment"], build_env_name)

        if "modules" in data and "variables" not in data:
            err_msg = (
                f"The 'modules' section is deprecated and has been renamed to 'variables' instead.\n"
                f"Please rename 'modules' to 'variables' in your config file., f{filepath.name}"
            )
            raise ToolkitEnvError(err_msg)
        variables = data.get("variables", {})
        return cls(environment=environment, variables=variables, filepath=filepath)

    def create_build_environment(
        self, built_modules: BuiltModuleList, selected_modules: ModuleDirectories
    ) -> BuildEnvironment:
        return BuildEnvironment(
            name=self.environment.name,  # type: ignore[arg-type]
            project=self.environment.project,
            validation_type=self.environment.validation_type,
            selected=self.environment.selected,
            cdf_toolkit_version=__version__,
            built_resources=built_modules.as_resources_by_folder(),
            read_modules=[module.as_read_module() for module in selected_modules],
        )

    def get_selected_modules(
        self,
        modules_by_package: dict[str, list[str | Path]],
        available_modules: set[str | Path],
        organization_dir: Path,
        verbose: bool,
    ) -> list[str | Path]:
        selected_packages = [
            package
            for package in self.environment.selected
            if package in modules_by_package and isinstance(package, str)
        ]
        if verbose:
            print("  [bold green]INFO:[/] Selected packages:")
            if len(selected_packages) == 0:
                print("    None")
            for package in selected_packages:
                print(f"    {package}")

        selected_modules = [module for module in self.environment.selected if module not in modules_by_package]
        if missing := set(selected_modules) - available_modules:
            hint = ModuleDefinition.long(missing, organization_dir)
            raise ToolkitMissingModuleError(
                f"The following selected modules are missing, please check path: {missing}.\n{hint}"
            )

        selected_modules.extend(
            itertools.chain.from_iterable(modules_by_package[package] for package in selected_packages)
        )
        if not selected_modules:
            raise ToolkitEnvError(
                f"No selected modules specified in {self.filepath!s}, have you configured "
                f"the environment ({self.environment.name})?"
            )
        if verbose:
            print("  [bold green]INFO:[/] Selected modules:")
            for module in selected_modules:
                if isinstance(module, Path):
                    print(f"    {module.as_posix()}")
                else:
                    print(f"    {module}")
        return selected_modules

    @classmethod
    def load_default(cls, organization_dir: Path) -> BuildConfigYAML:
        return cls(filepath=organization_dir / BuildConfigYAML.get_filename(DEFAULT_ENV))

    def dump(self) -> dict[str, Any]:
        return {"environment": self.environment.dump(), "variables": self.variables}


@dataclass
class BuildEnvironment(Environment):
    cdf_toolkit_version: str = __version__
    built_resources: dict[str, BuiltResourceList] = field(default_factory=dict)
    read_modules: list[ReadModule] = field(default_factory=list)

    @property
    def read_resource_folders(self) -> set[str]:
        return {resource_folder for module in self.read_modules for resource_folder in module.resource_directories}

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["cdf_toolkit_version"] = self.cdf_toolkit_version
        if self.built_resources:
            output["built_resources"] = {
                resource_folder: resources.dump(resource_folder, include_destination=True)
                for resource_folder, resources in self.built_resources.items()
            }
        if self.read_modules:
            output["read_modules"] = [module.dump() for module in self.read_modules]
        return output

    def dump_to_file(self, build_dir: Path) -> None:
        (build_dir / BUILD_ENVIRONMENT_FILE).write_text(
            "# DO NOT EDIT THIS FILE!\n" + yaml.dump(self.dump(), sort_keys=False, indent=2)
        )

    @classmethod
    def load(
        cls, data: dict[str, Any], build_name: str | None, action: Literal["build", "deploy", "clean", "pull"] = "build"
    ) -> BuildEnvironment:
        if "name" in data and build_name is not None and data["name"] != build_name:
            raise ToolkitEnvError(
                f"Expected to {action} for {build_name!r} environment, but the last "
                f"build was created for the {data['name']!r} environment."
            )
        build_name = build_name or data.get("name")

        version = _load_version_variable(data, BUILD_ENVIRONMENT_FILE)
        _deprecation_selected(data)
        built_resources: dict[str, BuiltResourceList] = {}
        if "built_resources" in data:
            # We expect to dump BuildEnvironment, and load DeployEnvironment
            built_resources = {
                resource_folder: BuiltResourceList.load(resources, resource_folder)
                for resource_folder, resources in data["built_resources"].items()
            }
        read_modules: list[ReadModule] = []
        if "read_modules" in data:
            read_modules = [ReadModule.load(module_data) for module_data in data["read_modules"]]
        _deprecate_type(data, build_name or "dev")
        try:
            return cls(
                name=data["name"],
                project=data["project"],
                validation_type=data["validation-type"],
                selected=data["selected"],
                cdf_toolkit_version=version,
                built_resources=built_resources,
                read_modules=read_modules,
            )
        except KeyError:
            raise ToolkitEnvError(
                f"  [bold red]ERROR:[/] Environment {build_name} is missing required fields 'name', 'project', 'validation-type', "
                f"or 'selected' in {BUILD_ENVIRONMENT_FILE!s}"
            )

    def set_environment_variables(self) -> None:
        os.environ["CDF_ENVIRON"] = self.name
        os.environ["CDF_BUILD_TYPE"] = self.validation_type

    def check_source_files_changed(self) -> WarningList[FileReadWarning]:
        warning_list = WarningList[FileReadWarning]()
        for resource_folder, resources in self.built_resources.items():
            if resource_folder == RawDatabaseLoader.folder_name:
                # We modify the hash for RawDatabaseLoader, so we skip checking the hash for this folder.
                continue
            for resource in resources:
                to_check = [resource.source, *(resource.extra_sources or [])]
                for source in to_check:
                    source_filepath = source.path
                    if source_filepath.suffix in {".csv", ".parquet"}:
                        # When we copy over the source files we use utf-8 encoding, which can change the file hash.
                        # Thus, we skip checking the hash for these file types.
                        continue

                    if not source_filepath.exists():
                        warning_list.append(
                            MissingFileWarning(source_filepath, attempted_check="source file has changed")
                        )
                    elif source.hash != calculate_hash(source_filepath, shorten=True):
                        warning_list.append(SourceFileModifiedWarning(source_filepath))
        return warning_list


def _deprecation_selected(data: dict[str, Any]) -> None:
    if "selected_modules_and_packages" in data and "selected" not in data:
        print(
            "  [bold yellow]Warning:[/] In environment section: 'selected_modules_and_packages' "
            "is deprecated, use 'selected' instead."
        )
        data["selected"] = data.pop("selected_modules_and_packages")


def _deprecate_type(data: dict[str, Any], build_name: str) -> None:
    if "type" in data and "validation-type" not in data:
        MediumSeverityWarning(
            f"In environment section of {BuildConfigYAML.get_filename(build_name)!s}: 'type' is deprecated, use 'validation-type' instead."
        ).print_warning()
        data["validation-type"] = data.pop("type")


@dataclass
class ConfigEntry:
    """This represents a single entry in a config.yaml file.

    Note that an entry is

    Args:
        key_path: The path to the variable, e.g. ('cognite_modules', 'another_module', 'source_asset')
        current_value: The current value of the variable
        default_value: The default value of the variable
        current_comment: The comment attached to the variable in the current config.yaml file
        default_comment: The comment attached to the variable in the default.config.yaml files in the module directories.
    """

    key_path: tuple[str, ...]
    current_value: float | int | str | bool | None = None
    default_value: float | int | str | bool | None = None
    current_comment: YAMLComment | None = None
    default_comment: YAMLComment | None = None

    @property
    def value(self) -> float | int | str | bool:
        if self.current_value is not None:
            return self.current_value
        elif self.default_value is not None:
            return self.default_value
        else:
            raise ValueError("config.yaml has not been loaded correctly, both default and current values are None")

    @property
    def comment(self) -> YAMLComment | None:
        return self.current_comment or self.default_comment

    @property
    def is_added(self) -> bool:
        return self.current_value is None and self.default_value is not None

    @property
    def is_removed(self) -> bool:
        return self.current_value is not None and self.default_value is None

    @property
    def is_unchanged(self) -> bool:
        # Default and current value can be different. Changed means
        # that the variable is either added or removed.
        return self.current_value is not None and self.default_value is not None

    def __str__(self) -> str:
        if self.is_removed:
            return f"{self.key} was removed"
        elif self.is_added:
            if len(self.key_path) < 2:
                return f"{self.key!r} was added with value {self.value!r}"
            variable = self.key_path[-1]
            module = ".".join(self.key_path[:-1])
            return f"{variable!r} was added to {module!r} with value {self.value!r}"
        else:
            return f"{self.key} is unchanged"

    def __repr__(self) -> str:
        return f"{self.key}={self.current_value!r}"

    @property
    def key(self) -> str:
        return ".".join(self.key_path)


class _WildcardSequence(tuple, Sequence[str]):
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Sequence)
            and len(other) == len(self)
            and all(a == b or a == "*" for a, b in zip(self, other))
        )

    def __hash__(self) -> int:
        return hash(tuple(self))


class WildcardSet(set, Set[_WildcardSequence]):
    """Sets that support wildcard matching."""

    @classmethod
    def load(cls, items: Sequence[Sequence[str]]) -> WildcardSet:
        return cls(_WildcardSequence(item) for item in items)

    def __contains__(self, key: Any) -> bool:
        if not isinstance(key, Sequence):
            return False
        return any(pattern == key for pattern in self)


@dataclass
class InitConfigYAML(YAMLWithComments[tuple[str, ...], ConfigEntry], ConfigYAMLCore):
    """This represents the 'config.[env].yaml' file in the root of the project.
    It is used in the init command.

    The motivation for having a specialist data structure and not just a dictionary:

    1. We want to be able to dump the config.yaml file with comments.
    2. We want to track which variables are added, removed, or unchanged.
    """

    # Top level keys
    _environment = "environment"
    _variables = "variables"

    def __init__(self, environment: Environment, entries: dict[tuple[str, ...], ConfigEntry] | None = None):
        self.environment = environment
        super().__init__(entries or {})

    def as_build_config(self) -> BuildConfigYAML:
        return BuildConfigYAML(environment=self.environment, variables=self.dump()[self._variables], filepath=Path(""))

    def load_defaults(
        self,
        cognite_root_module: Path,
        selected_paths: set[Path] | None = None,
        ignore_patterns: list[tuple[str, ...]] | None = None,
    ) -> InitConfigYAML:
        """Loads all default.config.yaml files in the cognite root module."""

        default_files_iterable: Iterable[Path]
        if cognite_root_module.name in ROOT_MODULES or cognite_root_module.name == BUILTIN_MODULES:
            default_files_iterable = cognite_root_module.glob(f"**/{DEFAULT_CONFIG_FILE}")
        else:
            default_files_iterable = itertools.chain(
                *[
                    (cognite_root_module / root_module).glob(f"**/{DEFAULT_CONFIG_FILE}")
                    for root_module in ROOT_MODULES
                    if (cognite_root_module / root_module).exists()
                ]
            )
        if selected_paths:
            default_files_iterable = (
                file
                for file in default_files_iterable
                if file.relative_to(cognite_root_module).parent in selected_paths
            )

        default_files = sorted(default_files_iterable, key=lambda f: f.relative_to(cognite_root_module))
        return self._load_defaults(cognite_root_module, default_files, ignore_patterns)

    def _load_defaults(
        self, cognite_root_module: Path, defaults_files: list[Path], ignore_patterns: list[tuple[str, ...]] | None
    ) -> InitConfigYAML:
        """Loads all default.config.yaml files in the cognite root module.

        This extracts the default values from the default.config.yaml files and
        adds them to the config.yaml file.

        Args:
            cognite_root_module: The root module for all cognite modules.

        Returns:
            self
        """

        ignore_set = WildcardSet.load(ignore_patterns) if ignore_patterns else None

        for default_config in defaults_files:
            parts = default_config.parent.relative_to(cognite_root_module).parts
            raw_file = safe_read(default_config)
            file_comments = self._extract_comments(raw_file, key_prefix=tuple(parts))
            file_data = cast(dict, read_yaml_content(raw_file))
            for key, value in file_data.items():
                if len(parts) >= 1 and parts[0] in ROOT_MODULES:
                    key_path = (self._variables, *parts, key)
                else:
                    key_path = (self._variables, MODULES, *parts, key)
                local_file_path = (*parts, key)
                if ignore_set and key_path[1:] in ignore_set:
                    continue
                if key_path in self:
                    self[key_path].default_value = value
                    self[key_path].default_comment = file_comments.get(local_file_path)
                else:
                    self[key_path] = ConfigEntry(
                        key_path=key_path,
                        default_value=value,
                        default_comment=file_comments.get(local_file_path),
                    )

        return self

    @classmethod
    def load_existing(cls, existing_config_yaml: str, build_env_name: str = "dev") -> InitConfigYAML:
        """Loads an existing config.yaml file.

        This does a yaml.safe_load, in addition to extracting comments from the file.

        Args:
            existing_config_yaml: The existing config.yaml file.
            build_env_name: The build environment.

        Returns:
            self

        """
        raw_file = existing_config_yaml
        comments = cls._extract_comments(raw_file)
        config = cast(dict, read_yaml_content(raw_file))
        if cls._environment in config:
            environment = Environment.load(config[cls._environment], build_env_name)
        else:
            raise ToolkitEnvError(f"Missing environment in {existing_config_yaml!s}")

        modules = config[cls._variables] if cls._variables in config else config
        entries: dict[tuple[str, ...], ConfigEntry] = {}
        for key_path, value in flatten_dict(modules).items():
            full_key_path = (cls._variables, *key_path)
            if full_key_path in entries:
                entries[full_key_path].current_value = value
                entries[full_key_path].current_comment = comments.get(full_key_path)
            else:
                entries[full_key_path] = ConfigEntry(
                    key_path=full_key_path,
                    current_value=value,
                    current_comment=comments.get(full_key_path),
                )

        return cls(
            environment=environment,
            entries=entries,
        )

    def load_variables(self, organization_dir: Path, propagate_reused_variables: bool = False) -> InitConfigYAML:
        """This scans the content the files in the given directories and finds the variables.
        The motivation is to find the variables that are used in the templates, as well
        as picking up variables that are used in custom modules.

        Variables are marked with a {{ variable }} syntax.

        Args:
            organization_dir: The directory with all project configurations.
            propagate_reused_variables: Whether to move variables with the same name to a shared parent.

        Returns:
            self
        """
        variable_by_parent_key: dict[str, set[tuple[str, ...]]] = defaultdict(set)
        for filepath in organization_dir.glob("**/*"):
            if filepath.suffix.lower() not in SEARCH_VARIABLES_SUFFIX:
                continue
            if filepath.name.startswith("default"):
                continue
            content = safe_read(filepath)
            key_parent = filepath.parent.relative_to(organization_dir).parts
            if key_parent and key_parent[-1] in LOADER_BY_FOLDER_NAME:
                key_parent = key_parent[:-1]

            for match in re.findall(r"{{\s*([a-zA-Z0-9_]+)\s*}}", content):
                variable_by_parent_key[match].add(key_parent)

        return self._load_variables(variable_by_parent_key, propagate_reused_variables)

    def _load_variables(
        self, variable_by_parent_key: dict[str, set[tuple[str, ...]]], propagate_reused_variables: bool = False
    ) -> InitConfigYAML:
        for variable, key_parents in variable_by_parent_key.items():
            if len(key_parents) > 1 and propagate_reused_variables:
                key_parents = {self._find_common_parent(list(key_parents))}

            for key_parent in key_parents:
                # Remove module subfolders.
                key_parent_list = list(key_parent)
                for i in range(len(key_parent_list)):
                    if key_parent_list[i] in LOADER_BY_FOLDER_NAME:
                        key_parent_list = key_parent_list[:i]
                        break
                key_parent = tuple(key_parent_list)
                key_path = (self._variables, *key_parent, variable)
                if key_path in self:
                    continue
                # Search for the first parent that match.
                for i in range(1, len(key_parent)):
                    alt_key_path = (self._variables, *key_parent[:i], variable)
                    if alt_key_path in self:
                        break
                else:
                    self[key_path] = ConfigEntry(key_path=key_path, current_value="<Not Set>")
        return self

    def lift(self) -> None:
        """Lift variables that are used in multiple modules to the highest shared level"""
        variables_by_key_value: dict[
            tuple[str, float | int | str | bool | tuple[Hashable] | None], list[ConfigEntry]
        ] = defaultdict(list)
        count_by_variable_keys: dict[str, set[float | int | str | bool | tuple[Hashable] | None]] = defaultdict(set)
        for key, entry in self.items():
            value = tuple(entry.value) if isinstance(entry.value, list) else entry.value  # type: ignore[arg-type]
            variables_by_key_value[(key[-1], value)].append(entry)
            count_by_variable_keys[key[-1]].add(value)

        for (variable_name, _), entries in variables_by_key_value.items():
            if len(entries) == 1:
                continue
            if len(count_by_variable_keys[variable_name]) > 1:
                continue
            shared_parent = self._find_common_parent([entry.key_path for entry in entries])
            new_key = (*shared_parent, entries[0].key_path[-1])
            self[new_key] = ConfigEntry(
                key_path=new_key, current_value=entries[0].current_value, default_value=entries[0].default_value
            )
            for entry in entries:
                del self[entry.key_path]

    @property
    def removed(self) -> list[ConfigEntry]:
        return [entry for entry in self.values() if entry.is_removed]

    @property
    def added(self) -> list[ConfigEntry]:
        return [entry for entry in self.values() if entry.is_added]

    @property
    def unchanged(self) -> list[ConfigEntry]:
        return [entry for entry in self.values() if entry.is_unchanged]

    def dump(self) -> dict[str, Any]:
        config: dict[str, Any] = {}
        for entry in self.values():
            local_config = config
            for key in entry.key_path[:-1]:
                if key not in local_config:
                    local_config[key] = {}
                local_config = local_config[key]
            local_config[entry.key_path[-1]] = entry.value
        config = self._reorder_config_yaml(config)
        return {
            self._environment: self.environment.dump(),
            **config,
        }

    def dump_yaml_with_comments(self, indent_size: int = 2) -> str:
        """Dump a config dictionary to a yaml string"""
        return self._dump_yaml_with_comments(indent_size, True)

    def _get_comment(self, key: tuple[str, ...]) -> YAMLComment | None:
        if (entry := self.get(key)) and entry.comment:
            return entry.comment
        return None

    def __str__(self) -> str:
        total_variables = len(self)
        lines = []
        if removed := self.removed:
            lines.append(f"Untracked {len(removed)} variables in {self.environment.name}.config.yaml.")
        if added := self.added:
            lines.append(f"Added {len(added)} variables to {self.environment.name}.config.yaml.")
        if total_variables == len(self.unchanged):
            lines.append(f"No variables in {self.environment.name}.config.yaml were changed.")
        return "\n".join(lines)

    @classmethod
    def _reorder_config_yaml(cls, config: dict[str, Any]) -> dict[str, Any]:
        """Reorder the config.yaml the variables before the modules."""
        new_config = {}
        for key in [k for k in config.keys() if not isinstance(config[k], dict)]:
            new_config[key] = config[key]
        for key in [k for k in config.keys() if isinstance(config[k], dict)]:
            new_config[key] = cls._reorder_config_yaml(config[key])
        return new_config

    @staticmethod
    def _find_common_parent(key_parents: list[tuple[str, ...]]) -> tuple[str, ...]:
        """Find the common parent for a list of key parents."""
        if len(key_parents) == 1:
            return key_parents[0]
        common_parent = []
        for i in range(len(key_parents[0])):
            if len({key_parent[i] if i < len(key_parent) else None for key_parent in key_parents}) == 1:
                common_parent.append(key_parents[0][i])
            else:
                break
        return tuple(common_parent)


class ConfigYAMLs(UserDict[str, InitConfigYAML]):
    def __init__(self, entries: dict[str, InitConfigYAML] | None = None):
        super().__init__(entries or {})

    @classmethod
    def load_default_environments(cls, default: dict[str, Any]) -> ConfigYAMLs:
        instance = cls()
        for environment_name, environment_config in default.items():
            environment = Environment.load(environment_config, environment_name)
            instance[environment.name] = InitConfigYAML(environment)
        return instance

    @classmethod
    def load_existing_environments(cls, existing_config_yamls: Sequence[Path]) -> ConfigYAMLs:
        instance = cls()
        for config_yaml in existing_config_yamls:
            config = InitConfigYAML.load_existing(safe_read(config_yaml), config_yaml.name.split(".")[0])
            instance[config.environment.name] = config
        return instance

    def load_default_variables(self, cognite_module: Path) -> None:
        # Can be optimized, but not a priority
        for config_yaml in self.values():
            config_yaml.load_defaults(cognite_module)

    def load_variables(self, organization_dir: Path) -> None:
        # Can be optimized, but not a priority
        for config_yaml in self.values():
            config_yaml.load_variables(organization_dir)



import sys
from types import MappingProxyType, UnionType
from typing import Any, ClassVar, Literal, cast, get_args

from pydantic import ModelWrapValidatorHandler, field_validator, model_serializer, model_validator
from pydantic_core.core_schema import SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from .base import BaseModelResource


class Scope(BaseModelResource):
    _scope_name: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_scope_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, Scope):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid scope data '{type(data)}' expected dict")

        if cls is not Scope:
            return handler(data)
        name, content = next(iter(data.items()))
        if name not in _SCOPE_CLASS_BY_NAME:
            raise ValueError(
                f"invalid scope name '{name}'. Expected one of {humanize_collection(_SCOPE_CLASS_BY_NAME.keys(), bind_word='or')}"
            )
        cls_ = _SCOPE_CLASS_BY_NAME[name]
        return cast(Self, cls_.model_validate(content))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_scope_name(self, handler: SerializerFunctionWrapHandler) -> dict:
        if self._scope_name is None:
            raise ValueError("Scope name is not set")
        serialized_data = handler(self)
        return {self._scope_name: serialized_data}


class AllScope(Scope):
    _scope_name = "all"


class CurrentUserScope(Scope):
    _scope_name = "currentuserscope"


class IDScope(Scope):
    _scope_name = "idScope"
    ids: list[str]


class IDScopeLowerCase(Scope):
    """Necessary due to lack of API standardisation on scope name: 'idScope' VS 'idscope'"""

    _scope_name = "idscope"
    ids: list[str]


class InstancesScope(Scope):
    _scope_name = "instancesScope"
    instances: list[str]


class ExtractionPipelineScope(Scope):
    _scope_name = "extractionPipelineScope"
    ids: list[str]


class PostgresGatewayUsersScope(Scope):
    _scope_name = "usersScope"
    usernames: list[str]


class DataSetScope(Scope):
    _scope_name = "datasetScope"
    ids: list[str]


class TableScope(Scope):
    _scope_name = "tableScope"
    dbs_to_tables: dict[str, list[str]]


class AssetRootIDScope(Scope):
    _scope_name = "assetRootIdScope"
    root_ids: list[str]


class ExperimentScope(Scope):
    _scope_name = "experimentscope"
    experiments: list[str]


class SpaceIDScope(Scope):
    _scope_name = "spaceIdScope"
    space_ids: list[str]


class PartitionScope(Scope):
    _scope_name = "partition"
    partition_ids: list[int]


class LegacySpaceScope(Scope):
    _scope_name = "spaceScope"
    external_ids: list[str]


class LegacyDataModelScope(Scope):
    _scope_name = "dataModelScope"
    external_ids: list[str]


class Capability(BaseModelResource):
    _capability_name: ClassVar[str]
    scope: Scope

    @field_validator("scope", mode="before")
    @classmethod
    def find_scope_cls(cls, data: Any) -> Scope:
        annotation = cls.model_fields["scope"].annotation
        if isinstance(annotation, UnionType):
            valid_types = {s._scope_name for s in get_args(annotation)}
        elif annotation is not None and issubclass(annotation, Scope):
            valid_types = {annotation._scope_name}
        else:
            raise ValueError(f"Invalid scope annotation '{annotation}'")

        name = next(iter(data.keys()))
        if name not in valid_types:
            raise ValueError(
                f"invalid scope name '{name}'. Expected {humanize_collection(valid_types, bind_word='or')}"
            )

        return Scope.model_validate(data)

    @model_validator(mode="wrap")
    @classmethod
    def find_capability_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if cls is not Capability:
            return handler(data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid capability data '{type(data)}' expected dict")
        name, content = next(iter(data.items()))
        if name not in _CAPABILITY_CLASS_BY_NAME:
            raise ValueError(f"Invalid capability name '{name}'. Expected one of {_CAPABILITY_CLASS_BY_NAME.keys()}")
        cls_ = _CAPABILITY_CLASS_BY_NAME[name]
        return cast(Self, cls_.model_validate(content))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_capability_name(self, handler: SerializerFunctionWrapHandler) -> dict:
        if self._capability_name is None:
            raise ValueError("Capability name is not set")
        serialized_data = handler(self)
        return {self._capability_name: serialized_data}


class AnalyticsAcl(Capability):
    _capability_name = "analyticsAcl"
    actions: list[Literal["READ", "EXECUTE", "LIST"]]
    scope: AllScope


class AnnotationsAcl(Capability):
    _capability_name = "annotationsAcl"
    actions: list[Literal["READ", "WRITE", "SUGGEST", "REVIEW"]]
    scope: AllScope


class AssetsAcl(Capability):
    _capability_name = "assetsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class DataSetsAcl(Capability):
    _capability_name = "datasetsAcl"
    actions: list[Literal["READ", "WRITE", "OWNER"]]
    scope: AllScope | IDScope


class DiagramParsingAcl(Capability):
    _capability_name = "diagramParsingAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class DigitalTwinAcl(Capability):
    _capability_name = "digitalTwinAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class EntityMatchingAcl(Capability):
    _capability_name = "entitymatchingAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class EventsAcl(Capability):
    _capability_name = "eventsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class ExtractionPipelinesAcl(Capability):
    _capability_name = "extractionPipelinesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | IDScope | DataSetScope


class ExtractionsRunAcl(Capability):
    _capability_name = "extractionRunsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | ExtractionPipelineScope


class ExtractionConfigsAcl(Capability):
    _capability_name = "extractionConfigsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | ExtractionPipelineScope


class FilesAcl(Capability):
    _capability_name = "filesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class FunctionsAcl(Capability):
    _capability_name = "functionsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class GeospatialAcl(Capability):
    _capability_name = "geospatialAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class GeospatialCrsAcl(Capability):
    _capability_name = "geospatialCrsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class GroupsAcl(Capability):
    _capability_name = "groupsAcl"
    actions: list[Literal["CREATE", "DELETE", "READ", "LIST", "UPDATE"]]
    scope: AllScope | CurrentUserScope


class LabelsAcl(Capability):
    _capability_name = "labelsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class LocationFiltersAcl(Capability):
    _capability_name = "locationFiltersAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | IDScope


class ProjectsAcl(Capability):
    _capability_name = "projectsAcl"
    actions: list[Literal["READ", "CREATE", "LIST", "UPDATE", "DELETE"]]
    scope: AllScope


class RawAcl(Capability):
    _capability_name = "rawAcl"
    actions: list[Literal["READ", "WRITE", "LIST"]]
    scope: AllScope | TableScope


class RelationshipsAcl(Capability):
    _capability_name = "relationshipsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class RoboticsAcl(Capability):
    _capability_name = "roboticsAcl"
    actions: list[Literal["READ", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | DataSetScope


class SAPWritebackAcl(Capability):
    _capability_name = "sapWritebackAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | InstancesScope


class SAPWritebackRequestsAcl(Capability):
    _capability_name = "sapWritebackRequestsAcl"
    actions: list[Literal["WRITE", "LIST"]]
    scope: AllScope | InstancesScope


class SecurityCategoriesAcl(Capability):
    _capability_name = "securityCategoriesAcl"
    actions: list[Literal["MEMBEROF", "LIST", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | IDScopeLowerCase


class SeismicAcl(Capability):
    _capability_name = "seismicAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | PartitionScope


class SequencesAcl(Capability):
    _capability_name = "sequencesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class SessionsAcl(Capability):
    _capability_name = "sessionsAcl"
    actions: list[Literal["LIST", "CREATE", "DELETE"]]
    scope: AllScope


class ThreeDAcl(Capability):
    _capability_name = "threedAcl"
    actions: list[Literal["READ", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | DataSetScope


class TimeSeriesAcl(Capability):
    _capability_name = "timeSeriesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | IDScopeLowerCase | AssetRootIDScope


class TimeSeriesSubscriptionsAcl(Capability):
    _capability_name = "timeSeriesSubscriptionsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TransformationsAcl(Capability):
    _capability_name = "transformationsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TypesAcl(Capability):
    _capability_name = "typesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class WellsAcl(Capability):
    _capability_name = "wellsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class ExperimentsAcl(Capability):
    _capability_name = "experimentAcl"
    actions: list[Literal["USE"]]
    scope: ExperimentScope


class TemplateGroupsAcl(Capability):
    _capability_name = "templateGroupsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TemplateInstancesAcl(Capability):
    _capability_name = "templateInstancesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class DataModelInstancesAcl(Capability):
    _capability_name = "dataModelInstancesAcl"
    actions: list[Literal["READ", "WRITE", "WRITE_PROPERTIES"]]
    scope: AllScope | SpaceIDScope


class DataModelsAcl(Capability):
    _capability_name = "dataModelsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | SpaceIDScope


class PipelinesAcl(Capability):
    _capability_name = "pipelinesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class DocumentPipelinesAcl(Capability):
    _capability_name = "documentPipelinesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class FilePipelinesAcl(Capability):
    _capability_name = "filePipelinesAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class NotificationsAcl(Capability):
    _capability_name = "notificationsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class ScheduledCalculationsAcl(Capability):
    _capability_name = "scheduledCalculationsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class MonitoringTasksAcl(Capability):
    _capability_name = "monitoringTasksAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class HostedExtractorsAcl(Capability):
    _capability_name = "hostedExtractorsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class VisionModelAcl(Capability):
    _capability_name = "visionModelAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class DocumentFeedbackAcl(Capability):
    _capability_name = "documentFeedbackAcl"
    actions: list[Literal["CREATE", "READ", "DELETE"]]
    scope: AllScope


class WorkflowOrchestrationAcl(Capability):
    _capability_name = "workflowOrchestrationAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class PostgresGatewayAcl(Capability):
    _capability_name = "postgresGatewayAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope | PostgresGatewayUsersScope


class UserProfilesAcl(Capability):
    _capability_name = "userProfilesAcl"
    actions: list[Literal["READ"]]
    scope: AllScope


class AuditlogAcl(Capability):
    _capability_name = "auditlogAcl"
    actions: list[Literal["READ"]]
    scope: AllScope


class VideoStreamingAcl(Capability):
    _capability_name = "videoStreamingAcl"
    actions: list[Literal["READ", "WRITE", "SUBSCRIBE", "PUBLISH"]]
    scope: AllScope | DataSetScope


class LegacyModelHostingAcl(Capability):
    _capability_name = "modelHostingAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


class LegacyGenericsAcl(Capability):
    _capability_name = "genericsAcl"
    actions: list[Literal["READ", "WRITE"]]
    scope: AllScope


_CAPABILITY_CLASS_BY_NAME: MappingProxyType[str, type[Capability]] = MappingProxyType(
    {c._capability_name: c for c in Capability.__subclasses__()}
)
ALL_CAPABILITIES = sorted(_CAPABILITY_CLASS_BY_NAME)

_SCOPE_CLASS_BY_NAME: MappingProxyType[str, type[Scope]] = MappingProxyType(
    {s._scope_name: s for s in Scope.__subclasses__()}
)

_SCOPE_BY_CLASS_NAME: MappingProxyType[str, str] = MappingProxyType(
    {scope_cls.__name__: scope_name for scope_name, scope_cls in _SCOPE_CLASS_BY_NAME.items()}
)



from datetime import datetime, timedelta, timezone
from itertools import islice
from timeit import default_timer

from cognite.client import CogniteClient
from cognite.client.data_classes import ExtractionPipelineRun
from cognite.client.data_classes.data_modeling import NodeId, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteAsset, CogniteTimeSeries
from cognite.client.data_classes.filters import Prefix, ContainsAny

from ice_cream_factory_api import IceCreamFactoryAPI

from cognite.client.config import global_config
global_config.disable_pypi_version_check = True

from itertools import islice


def batcher(iterable, batch_size):
    iterator = iter(iterable)
    while batch := list(islice(iterator, batch_size)):
        yield batch


def get_time_series_for_site(client: CogniteClient, site):
    this_site = site.lower()
    sub_tree_root = client.data_modeling.instances.retrieve_nodes(
        NodeId("icapi_dm_space", this_site),
        node_cls=CogniteAsset
    )

    if not sub_tree_root:
        print(
            f"----No CogniteAssets in CDF for {site}!----\n"
            f"    Run the 'Create Cognite Asset Hierarchy' transformation!"
        )
        return []

    sub_tree_nodes = client.data_modeling.instances.list(
        instance_type=CogniteAsset,
        filter=Prefix(property=["cdf_cdm", "CogniteAsset/v1", "path"], value=sub_tree_root.path),
        limit=None
    )

    if not sub_tree_nodes:
        print(
            f"----No CogniteTimeSeries in CDF for {site}!----\n"
            f"    Run the 'Contextualize Timeseries and Assets' transformation!"
        )
        return []

    value_list = [{"space": node.space, "externalId": node.external_id} for node in sub_tree_nodes]

    time_series = [
        client.data_modeling.instances.search(
            view=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
            instance_type=CogniteTimeSeries,
            filter=ContainsAny(property=["cdf_cdm", "CogniteTimeSeries/v1", "assets"], values=batch),
            limit=None
        )
        for batch in batcher(value_list, 20)
    ]

    # Combine list of batch results into a single NodeList
    time_series = [node for nodelist in time_series for node in nodelist]

    if not time_series:
        print("No CogniteTimeSeries in the CogniteCore Data Model (cdf_cdm Space)")

    time_series = [
        item for item in time_series
        if any(substring in item.external_id for substring in ["planned_status", "good"])
    ]

    return time_series


def report_ext_pipe(client: CogniteClient, status, message=None):
    ext_pipe_run = ExtractionPipelineRun(
        extpipe_external_id="ep_icapi_datapoints",
        status=status,
        message=message
    )

    client.extraction_pipelines.runs.create(run=ext_pipe_run)

def handle(client: CogniteClient = None, data=None):
    report_ext_pipe(client, "seen")
    
    sites = None
    backfill = None
    hours = None
    max_hours = 336

    if data:
        sites = data.get("sites")
        backfill = data.get("backfill")
        hours = data.get("hours")

        if hours and hours > max_hours:
            print(f"{hours} > {max_hours}! The Ice Cream API can't serve more than {max_hours} hours of datapoints, setting hours to max")
            hours = max_hours

    all_sites = [
        "Houston",
        "Oslo",
        "Kuala_Lumpur",
        "Hannover",
        "Nuremberg",
        "Marseille",
        "Sao_Paulo",
        "Chicago",
        "Rotterdam",
        "London",
    ]

    sites = sites or all_sites
    backfill = backfill or True
    hours = hours or max_hours

    now = datetime.now(timezone.utc).timestamp() * 1000
    increment = timedelta(hours=hours).total_seconds() * 1000

    ice_cream_api = IceCreamFactoryAPI(base_url="https://ice-cream-factory.inso-internal.cognite.ai")

    try:
        for site in sites:
            print(f"Getting Data Points for {site}")
            big_start = default_timer()

            time_series = get_time_series_for_site(client, site)

            latest_dps = {
                dp.external_id: dp.timestamp
                for dp in client.time_series.data.retrieve_latest(
                    external_id=[ts.external_id for ts in time_series],
                    ignore_unknown_ids=True
                )
            } if not backfill else None

            to_insert = []
            for ts in time_series:
                # figure out the window of datapoints to pull for this Time Series
                latest = latest_dps[ts.external_id][0] if not backfill and latest_dps.get(ts.external_id) else None

                start = latest if latest else now - increment
                end = now
            
                dps_list = ice_cream_api.get_datapoints(timeseries_ext_id=ts.external_id, start=start, end=end)

                for dp_dict in dps_list:
                    dp_dict["instance_id"] = NodeId(space="icapi_dm_space", external_id=dp_dict["instance_id"])

                to_insert.extend(dps_list)

                if len(to_insert) > 50:
                    client.time_series.data.insert_multiple(datapoints=to_insert)
                    to_insert = []

            if to_insert:
                client.time_series.data.insert_multiple(datapoints=to_insert)
                print(f"  {hours}h of Datapoints took {default_timer() - big_start:.2f} seconds")
            else:
                print(f"  No TimeSeries, for {hours}h of Datapoints took {default_timer() - big_start:.2f} seconds")

        report_ext_pipe(client, "success")
    except Exception as e:
        report_ext_pipe(client, "fail", e)


import shutil
from importlib import resources
from pathlib import Path

import questionary

import cognite_toolkit
from cognite_toolkit._cdf_tk.constants import REPO_FILES_DIR
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning, MediumSeverityWarning

from . import _cli_commands
from ._base import ToolkitCommand

REPOSITORY_HOSTING = [
    "GitHub",
    "Azure DevOps",
    "Other",
    "None",
]


class RepoCommand(ToolkitCommand):
    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
        skip_git_verify: bool = False,
    ) -> None:
        super().__init__(print_warning=print_warning, skip_tracking=skip_tracking, silent=silent)
        self._repo_files = Path(resources.files(cognite_toolkit.__name__)) / REPO_FILES_DIR  # type: ignore [arg-type]
        self.skip_git_verify = skip_git_verify

    def init(self, cwd: Path, host: str | None = None, verbose: bool = False) -> None:
        if not self.skip_git_verify:
            if _cli_commands.use_git():
                if not _cli_commands.has_initiated_repo():
                    self.warn(MediumSeverityWarning("git repository not initiated. Did you forget to run `git init`?"))
                else:
                    git_root = _cli_commands.git_root()
                    if git_root is None:
                        self.warn(MediumSeverityWarning("Unknown error when trying to find git root."))
                    elif cwd != git_root:
                        raise ToolkitValueError(
                            f"Current working directory is not the root of the git repository. "
                            f"Please run this command from {git_root.as_posix()!r}."  # type: ignore [union-attr]
                        )
            else:
                self.warn(
                    MediumSeverityWarning("git is not installed. It is strongly recommended to use version control.")
                )

        if host is None:
            repo_host = questionary.select("Where do are you hosting the repository?", REPOSITORY_HOSTING).ask()
        else:
            repo_host = next(
                (provider for provider in REPOSITORY_HOSTING if provider.casefold() == host.casefold()), "Other"
            )
        if repo_host == "GitHub":
            self.console("The repository will be hosted on GitHub.")
        elif repo_host == "Azure DevOps":
            self.console("The repository will be hosted on Azure DevOps.")
        elif repo_host == "Other":
            self.console("No template for CI/CD available for other hosting services yet.")
        elif repo_host == "None":
            self.console("It is recommended to use a hosted version control service like GitHub or Azure DevOps.")

        if verbose:
            self.console("Initializing toolkit repository...")

        iterables = [(self._repo_files, self._repo_files.glob("*"))]
        if repo_host in ["GitHub", "Azure DevOps"]:
            repo_host = repo_host.replace(" ", "")
            iterables.append((self._repo_files / repo_host, self._repo_files.rglob(f"{repo_host}/**/*.*")))

        for root, iterable in iterables:
            for file in iterable:
                if file.is_dir():
                    continue
                destination = cwd / file.relative_to(root)
                if destination.exists():
                    self.warn(LowSeverityWarning(f"File {destination} already exists. Skipping..."))
                    continue
                destination.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(file, destination)
                if verbose:
                    self.console(f"Created {destination.relative_to(cwd).as_posix()!r}")

        self.console("Repo initialization complete.")



import os
import re
import sys
from pathlib import Path

from typing_extensions import Literal, TypeAlias

try:
    from pyodide.ffi import IN_BROWSER  # type: ignore [import-not-found]
except ModuleNotFoundError:
    IN_BROWSER = False
# This is the default config located locally in each module.
# The environment file:

TOOLKIT_SERVICE_PRINCIPAL_GROUP_NAME = "cognite_toolkit_service_principal"
TOOLKIT_DEMO_GROUP_NAME = "cognite_toolkit_demo"

# This is the default Cognite app registration for Entra with device code enabled
# to be used with the Toolkit.
TOOLKIT_CLIENT_ENTRA_ID = "fb9d503b-ac25-44c7-a75d-8fbcd3a206bd"

_RUNNING_IN_BROWSER = IN_BROWSER
# This is the default config located locally in each module.
DEFAULT_CONFIG_FILE = "default.config.yaml"
# The environment file:
BUILD_ENVIRONMENT_FILE = "_build_environment.yaml"
# The local config file:
CONFIG_FILE_SUFFIX = "config.yaml"
# The global config file
GLOBAL_CONFIG_FILE = "global.yaml"

BUILTIN_MODULES = "_builtin_modules"
COGNITE_MODULES = "cognite_modules"
CUSTOM_MODULES = "custom_modules"
MODULES = "modules"
REPO_FILES_DIR = "_repo_files"
DOCKER_IMAGE_NAME = "cognite/toolkit"
BUILD_FOLDER_ENCODING = "utf-8"

ROOT_MODULES = [MODULES, CUSTOM_MODULES, COGNITE_MODULES]
MODULE_PATH_SEP = "/"

MIN_TIMESTAMP_MS = -2208988800000  # 1900-01-01 00:00:00.000
MAX_TIMESTAMP_MS = 4102444799999  # 2099-12-31 23:59:59.999
HAS_DATA_FILTER_LIMIT = 10

DEV_ONLY_MODULES = frozenset(["cdf_auth_readwrite_all"])

DEFAULT_ENV = "dev"
# Add any other files below that should be included in a build
EXCL_FILES = ["README.md", DEFAULT_CONFIG_FILE]
# Files to search for variables.
SEARCH_VARIABLES_SUFFIX = frozenset([".yaml", "yml", ".sql", ".csv"])
YAML_SUFFIX = frozenset([".yaml", ".yml"])
# Which files to process for template variable replacement
TEMPLATE_VARS_FILE_SUFFIXES = frozenset([".yaml", ".yml", ".sql", ".json", ".graphql"])
TABLE_FORMATS = frozenset([".csv", ".parquet"])
ROOT_PATH = Path(__file__).parent.parent
COGNITE_MODULES_PATH = ROOT_PATH / COGNITE_MODULES
MODULES_PATH = ROOT_PATH / MODULES
BUILTIN_MODULES_PATH = ROOT_PATH / BUILTIN_MODULES
SUPPORT_MODULE_UPGRADE_FROM_VERSION = "0.2.0"
# This is used in the build directory to keep track of order and flatten the
# module directory structure with accounting for duplicated names.
INDEX_PATTERN = re.compile("^[0-9]+\\.")
# This is used in the lookup of for example datasets to indicate that is a dry run.
DRY_RUN_ID = -1

# This is a regular expression that matches any non-word character or underscore
# It is used to clean the feature flag names.
_CLEAN_PATTERN = re.compile(r"[\W_]+")

# This is used to detect environment variables in a string.
ENV_VAR_PATTERN = re.compile(r"\$\{([^}]+)\}")

HINT_LEAD_TEXT = "[bold blue]HINT[/bold blue] "
HINT_LEAD_TEXT_LEN = 5
EnvType: TypeAlias = Literal["dev", "test", "staging", "qa", "prod"]
USE_SENTRY = "pytest" not in sys.modules and os.environ.get("SENTRY_ENABLED", "true").lower() == "true"
SPACE_FORMAT_PATTERN = r"^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$"
CONTAINER_EXTERNAL_ID_PATTERN = r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$"
FORBIDDEN_SPACES = frozenset(["space", "cdf", "dms", "pg3", "shared", "system", "node", "edge"])
FORBIDDEN_CONTAINER_EXTERNAL_IDS = frozenset(
    [
        "Query",
        "Mutation",
        "Subscription",
        "String",
        "Int32",
        "Int64",
        "Int",
        "Float32",
        "Float64",
        "Float",
        "Timestamp",
        "JSONObject",
        "Date",
        "Numeric",
        "Boolean",
        "PageInfo",
        "File",
        "Sequence",
        "TimeSeries",
    ]
)
FORBIDDEN_CONTAINER_PROPERTIES_IDENTIFIER = frozenset(
    [
        "space",
        "externalId",
        "createdTime",
        "lastUpdatedTime",
        "deletedTime",
        "edge_id",
        "node_id",
        "project_id",
        "property_group",
        "seq",
        "tg_table_name",
        "extensions",
    ]
)
CONTAINER_PROPERTIES_IDENTIFIER_PATTERN = r"^[a-zA-Z0-9][a-zA-Z0-9_-]{0,253}[a-zA-Z0-9]?$"


def clean_name(name: str) -> str:
    """Cleans the name by removing any non-word characters or underscores."""
    return _CLEAN_PATTERN.sub("", name).casefold()


class URL:
    configure_access = "https://docs.cognite.com/cdf/deploy/cdf_deploy/cdf_deploy_access_management"
    auth_toolkit = "https://docs.cognite.com/cdf/deploy/cdf_toolkit/guides/configure_deploy_modules#configure-the-cdf-toolkit-authentication"
    docs = "https://docs.cognite.com/"
    configs = "https://docs.cognite.com/cdf/deploy/cdf_toolkit/references/configs"
    plugins = "https://docs.cognite.com/cdf/deploy/cdf_toolkit/guides/plugins/"
    libyaml = "https://pyyaml.org/wiki/PyYAMLDocumentation"
    build_variables = "https://docs.cognite.com/cdf/deploy/cdf_toolkit/api/config_yaml#the-variables-section"


# The number of instances that should be left as a margin when Toolkit writes to CDF through the DMS API.
# This is currently set conservatively to 1 million. The reasoning for this is that there are CDF
# applications such as Infield and Industrial Canvas that can write to the DMS API, as well as likely third-party
# applications that can write to the DMS API. If Toolkit fills up the entire capacity, these type of data gathering
# applications will experience data loss. The limit of 1 million is chosen such that it will trigger alarms in the
# CDF projects, such that admins can take action to increase or clean up the capacity before it is too late.
DMS_INSTANCE_LIMIT_MARGIN = 1_000_000



from __future__ import annotations

import pytest
from cognite.client.data_classes.data_modeling import ContainerApply, SpaceApply

from cognite_toolkit._cdf_tk._parameters import (
    ANY_INT,
    ANY_STR,
    ANYTHING,
    ParameterSet,
    ParameterSpec,
    ParameterSpecSet,
    ParameterValue,
    read_parameter_from_init_type_hints,
    read_parameters_from_dict,
)
from cognite_toolkit._cdf_tk.loaders import RESOURCE_LOADER_LIST, ResourceLoader
from cognite_toolkit._cdf_tk.loaders._resource_loaders.location_loaders import LocationFilterLoader


class TestReadParameterFromTypeHints:
    @pytest.mark.parametrize(
        "cls_, expected_parameters",
        [
            (
                SpaceApply,
                ParameterSpecSet(
                    {
                        ParameterSpec(("space",), frozenset({"str"}), True, False),
                        ParameterSpec(("description",), frozenset({"str"}), False, True),
                        ParameterSpec(("name",), frozenset({"str"}), False, True),
                    }
                ),
            ),
            (
                ContainerApply,
                ParameterSpecSet(
                    {
                        ParameterSpec(("space",), frozenset({"str"}), True, False),
                        ParameterSpec(("external_id",), frozenset({"str"}), True, False),
                        ParameterSpec(("properties",), frozenset({"dict"}), True, False),
                        ParameterSpec(
                            path=("properties", ANY_STR),
                            types=frozenset({"dict"}),
                            is_required=False,
                            _is_nullable=False,
                        ),
                        ParameterSpec(("properties", ANY_STR, "type"), frozenset({"dict"}), True, False),
                        ParameterSpec(("properties", ANY_STR, "nullable"), frozenset({"bool"}), False, False),
                        ParameterSpec(("properties", ANY_STR, "auto_increment"), frozenset({"bool"}), False, False),
                        ParameterSpec(("properties", ANY_STR, "immutable"), frozenset({"bool"}), False, False),
                        ParameterSpec(("properties", ANY_STR, "name"), frozenset({"str"}), False, True),
                        ParameterSpec(
                            ("properties", ANY_STR, "default_value"),
                            frozenset({"float", "bool", "str", "int", "dict"}),
                            False,
                            True,
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "default_value", ANYTHING), frozenset({"unknown"}), False, True
                        ),
                        ParameterSpec(("properties", ANY_STR, "description"), frozenset({"str"}), False, True),
                        ParameterSpec(("properties", ANY_STR, "type", "collation"), frozenset({"str"}), False, False),
                        ParameterSpec(("properties", ANY_STR, "type", "container"), frozenset({"dict"}), False, True),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "container", "space"), frozenset({"str"}), True, False
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "container", "external_id"), frozenset({"str"}), True, False
                        ),
                        ParameterSpec(("properties", ANY_STR, "type", "is_list"), frozenset({"bool"}), False, False),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "max_list_size"), frozenset({"int"}), False, True
                        ),
                        ParameterSpec(("properties", ANY_STR, "type", "unit"), frozenset({"dict"}), False, True),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "unit", "external_id"), frozenset({"str"}), True, False
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "unit", "source_unit"), frozenset({"str"}), False, True
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "unknown_value"), frozenset({"str"}), False, True
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "values"),
                            frozenset({"dict"}),
                            True,
                            False,
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "values", ANY_STR), frozenset({"dict"}), False, False
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "values", ANY_STR, "description"),
                            frozenset({"str"}),
                            False,
                            True,
                        ),
                        ParameterSpec(
                            ("properties", ANY_STR, "type", "values", ANY_STR, "name"), frozenset({"str"}), True, False
                        ),
                        ParameterSpec(("description",), frozenset({"str"}), False, True),
                        ParameterSpec(("name",), frozenset({"str"}), False, True),
                        ParameterSpec(("used_for",), frozenset({"str"}), False, True),
                        ParameterSpec(("constraints",), frozenset({"dict"}), False, True),
                        ParameterSpec(
                            path=("constraints", ANY_STR),
                            types=frozenset({"dict"}),
                            is_required=False,
                            _is_nullable=False,
                        ),
                        ParameterSpec(
                            (
                                "constraints",
                                ANY_STR,
                                "require",
                            ),
                            frozenset({"dict"}),
                            True,
                            False,
                        ),
                        ParameterSpec(("constraints", ANY_STR, "require", "space"), frozenset({"str"}), True, False),
                        ParameterSpec(
                            ("constraints", ANY_STR, "require", "external_id"), frozenset({"str"}), True, False
                        ),
                        ParameterSpec(("constraints", ANY_STR, "properties"), frozenset({"list"}), True, False),
                        ParameterSpec(
                            ("constraints", ANY_STR, "properties", ANY_INT), frozenset({"str"}), False, False
                        ),
                        ParameterSpec(("indexes",), frozenset({"dict"}), False, True),
                        ParameterSpec(
                            path=("indexes", ANY_STR), types=frozenset({"dict"}), is_required=False, _is_nullable=False
                        ),
                        ParameterSpec(("indexes", ANY_STR, "properties"), frozenset({"list"}), True, False),
                        ParameterSpec(("indexes", ANY_STR, "properties", ANY_INT), frozenset({"str"}), False, False),
                        ParameterSpec(("indexes", ANY_STR, "cursorable"), frozenset({"bool"}), False, False),
                    }
                ),
            ),
        ],
    )
    def test_read_as_expected(self, cls_: type, expected_parameters: ParameterSpecSet) -> None:
        actual_parameters = read_parameter_from_init_type_hints(cls_)

        assert sorted(actual_parameters) == sorted(expected_parameters)

    @pytest.mark.parametrize("loader_cls", RESOURCE_LOADER_LIST)
    def test_compatible_with_loaders(self, loader_cls: type[ResourceLoader]) -> None:
        if loader_cls is LocationFilterLoader:
            # TODO: https://cognitedata.atlassian.net/browse/CDF-22363
            pytest.skip(f"Skipping {loader_cls} because get_write_cls_parameter_spec fails for some reason")

        parameter_set = read_parameter_from_init_type_hints(loader_cls.resource_write_cls)

        assert isinstance(parameter_set, ParameterSpecSet)
        assert len(parameter_set) > 0


class TestReadParameterFromDict:
    @pytest.mark.parametrize(
        "raw, expected",
        [
            (
                {"space": "space", "description": "description", "name": "name"},
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("description",), "str", "description"),
                        ParameterValue(("name",), "str", "name"),
                    }
                ),
            ),
            (
                {
                    "externalId": "Asset",
                    "name": "Asset",
                    "space": "sp_asset_space",
                    "usedFor": "node",
                    "properties": {
                        "metadata": {
                            "type": {
                                "list": False,
                                "type": "json",
                            },
                            "nullable": False,
                            "autoIncrement": False,
                            "name": "name",
                            "defaultValue": "default_value",
                            "description": "description",
                        }
                    },
                },
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(("externalId",), "str", "Asset"),
                        ParameterValue(("name",), "str", "Asset"),
                        ParameterValue(("space",), "str", "sp_asset_space"),
                        ParameterValue(("usedFor",), "str", "node"),
                        ParameterValue(("properties",), "dict", None),
                        ParameterValue(
                            (
                                "properties",
                                "metadata",
                            ),
                            "dict",
                            None,
                        ),
                        ParameterValue(("properties", "metadata", "type"), "dict", None),
                        ParameterValue(("properties", "metadata", "type", "list"), "bool", False),
                        ParameterValue(("properties", "metadata", "type", "type"), "str", "json"),
                        ParameterValue(("properties", "metadata", "nullable"), "bool", False),
                        ParameterValue(("properties", "metadata", "autoIncrement"), "bool", False),
                        ParameterValue(("properties", "metadata", "name"), "str", "name"),
                        ParameterValue(("properties", "metadata", "defaultValue"), "str", "default_value"),
                        ParameterValue(("properties", "metadata", "description"), "str", "description"),
                    }
                ),
            ),
            ({}, ParameterSet[ParameterValue]({})),
        ],
    )
    def test_read_expected(self, raw: dict, expected: ParameterSet[ParameterValue]):
        actual = read_parameters_from_dict(raw)

        assert sorted(actual) == sorted(expected)



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import ThreeDModelYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_3D_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "MyModel"},
        {"Missing required field: 'name'", "Unused field: 'externalId'"},
        id="Missing required field: externalId",
    )
    yield pytest.param(
        {"name": "MyModel", "dataSetId": 123},
        {"Unused field: 'dataSetId'"},
        id="Unused field: dataSetId",
    )
    yield pytest.param(
        {"name": "MyModel", "metadata": {f"key{i}": f"value{i}" for i in range(17)}},
        {"In field metadata dictionary should have at most 16 items after validation, not 17"},
        id="Invalid metadata types",
    )


class TestDataSetYAML:
    @pytest.mark.parametrize("data", list(find_resources("3DModel")))
    def test_load_valid_dataset(self, data: dict[str, object]) -> None:
        loaded = ThreeDModelYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_3D_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, ThreeDModelYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from __future__ import annotations

import typer

from cognite_toolkit._cdf_tk.commands import InitCommand


class LandingApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.command()(self.main_init)

    def main_init(self) -> None:
        """Guidance on how to get started"""
        cmd = InitCommand()
        cmd.run(cmd.execute)



from .base import ToolkitResource


class TimeSeriesYAML(ToolkitResource):
    external_id: str
    name: str | None = None
    is_string: bool = False
    metadata: dict[str, str] | None = None
    unit: str | None = None
    unit_external_id: str | None = None
    asset_external_id: str | None = None
    is_step: bool = False
    description: str | None = None
    security_categories: list[str] | None = None
    data_set_external_id: str | None = None



from tests.test_unit.approval_client import ApprovalToolkitClient


class TestLookUpAPI:
    def test_lookup_dataset(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        external_id = "SomeDataSet"
        client = toolkit_client_approval.mock_client
        data_set_id = client.lookup.data_sets.id(external_id)

        recreated = client.lookup.data_sets.external_id(data_set_id)

        assert external_id == recreated



import fnmatch
from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import lru_cache
from pathlib import Path
from typing import ClassVar, Literal
from urllib.parse import urljoin

import requests
from requests import Response
from rich import print
from rich.progress import track

from cognite_toolkit._cdf_tk.constants import IN_BROWSER
from cognite_toolkit._cdf_tk.tk_warnings import HTTPWarning


class FileDownloader(ABC):
    _type: ClassVar[str] = "generic"

    @abstractmethod
    def __init__(self, repo: str, errors: Literal["continue", "raise"] = "continue") -> None:
        pass

    @abstractmethod
    def copy(self, source: str, destination: Path) -> None:
        pass


class GitHubFileDownloader(FileDownloader):
    _type: ClassVar[str] = "github"

    api_url = "https://api.github.com"

    def __init__(self, repo: str, errors: Literal["continue", "raise"] = "continue") -> None:
        self.repo = repo
        self.errors = errors

    def copy(self, source: str, destination: Path) -> None:
        source_path = Path(source)
        to_download = list(self._find_files(source_path))
        if IN_BROWSER:
            iterable = to_download
        else:
            iterable = track(to_download, description=f"Downloading from {source_path.as_posix()!r}")  # type: ignore [assignment]
        for path, url in iterable:
            self._download_file(url, path, destination)

    def _find_files(self, source: Path) -> Iterable[tuple[Path, str]]:
        search = [""]
        parents = set(source.parents)
        while search:
            path = search.pop()
            content = self._get_repo_contents(path)
            for item in content:
                repo_path = Path(item["path"])
                if item["type"] == "file" and fnmatch.fnmatch(repo_path.name, source.name):
                    yield repo_path, item["download_url"]
                elif item["type"] == "dir" and repo_path in parents:
                    search.append(item["path"])

    @lru_cache
    def _get_repo_contents(self, path: str = "") -> list[dict[str, str]]:
        url = f"{self.api_url}/repos/{self.repo}/contents/{path}"
        response = self._requests(url, "get repo contents")
        if response is None:
            return []
        return response.json()

    def _requests(self, url: str, action: str) -> Response | None:
        response = requests.get(url)
        if response.status_code >= 400:
            if self.errors == "raise":
                response.raise_for_status()
            print(HTTPWarning(action, response.text, response.status_code).get_message())
            return None
        return response

    def _download_file(self, url: str, source: Path, destination: Path) -> None:
        response = self._requests(url, f"download {source}")
        if response is None:
            return
        if destination.suffix:
            destination_path = destination
        else:
            destination_path = destination / source.name
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        destination_path.write_bytes(response.content)


class HttpFileDownloader(FileDownloader):
    _type: ClassVar[str] = "http"

    def __init__(self, repo: str, errors: Literal["continue", "raise"] = "raise") -> None:
        self.repo = repo
        self.errors = errors

    def copy(self, source: str, destination: Path) -> None:
        if not self.repo.endswith("/"):
            self.repo += "/"

        sources = source.split(";") if ";" in source else [source]
        for source in sources:
            file_url = urljoin(self.repo, source)

            response = requests.get(file_url)
            if response.status_code >= 400:
                if self.errors == "raise":
                    response.raise_for_status()
                print(HTTPWarning("GET", response.text, response.status_code).get_message())
                continue

            location = destination if not destination.is_dir() else destination / Path(source).name
            location.write_bytes(response.content)



import sys
from types import MappingProxyType
from typing import Any, ClassVar, Literal, cast

from pydantic import Field, ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import BaseModelResource

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class DataModelInfo(BaseModelResource):
    space: str = Field(description="Space of the Data Model.")
    external_id: str = Field(description="External ID of the Data Model.")
    version: str = Field(description="Version of the Data Model.")
    destination_type: str = Field(description="External ID of the type(view) in the data model.")
    destination_relationship_from_type: str | None = Field(
        default=None, description="Property Identifier of the connection definition in destinationType."
    )


class ViewInfo(BaseModelResource):
    space: str = Field(description="Space of the view.")
    external_id: str = Field(description="External ID of the view.")
    version: str = Field(description="Version of the view.")


class EdgeType(BaseModelResource):
    space: str = Field(description="Space of the type.")
    external_id: str = Field(description="External ID of the type.")


class Destination(BaseModelResource):
    _destination_type: ClassVar[str]
    type: str

    @model_validator(mode="wrap")
    @classmethod
    def find_destination_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, Destination):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid destination data '{type(data)}' expected dict")

        if cls is not Destination:
            return handler(data)

        dest_type = data.get("type")
        if dest_type is None:
            raise ValueError("Missing 'type' field in destination data")
        if dest_type not in _DESTINATION_CLASS_BY_TYPE:
            raise ValueError(
                f"invalid destination type '{dest_type}'. Expected one of {humanize_collection(_DESTINATION_CLASS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _DESTINATION_CLASS_BY_TYPE[dest_type]
        return cast(Self, cls_.model_validate(data))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def serialize_destination(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        return serialized_data


class StandardDataSource(Destination):
    _destination_type: ClassVar[str]
    type: Literal[
        "assets",
        "events",
        "asset_hierarchy",
        "datapoints",
        "string_datapoints",
        "timeseries",
        "sequences",
        "files",
        "labels",
        "relationships",
        "data_sets",
    ]


class AssetsDataSource(StandardDataSource):
    _destination_type = "assets"
    type: Literal["assets"] = "assets"


class EventsDataSource(StandardDataSource):
    _destination_type = "events"
    type: Literal["events"] = "events"


class AssetHierarchyDataSource(StandardDataSource):
    _destination_type = "asset_hierarchy"
    type: Literal["asset_hierarchy"] = "asset_hierarchy"


class DataPointsDataSource(StandardDataSource):
    _destination_type = "datapoints"
    type: Literal["datapoints"] = "datapoints"


class StringDataPointsDataSource(StandardDataSource):
    _destination_type = "string_datapoints"
    type: Literal["string_datapoints"] = "string_datapoints"


class TimeSeriesDataSource(StandardDataSource):
    _destination_type = "timeseries"
    type: Literal["timeseries"] = "timeseries"


class SequencesDataSource(StandardDataSource):
    _destination_type = "sequences"
    type: Literal["sequences"] = "sequences"


class FilesDataSource(StandardDataSource):
    _destination_type = "files"
    type: Literal["files"] = "files"


class LabelsDataSource(StandardDataSource):
    _destination_type = "labels"
    type: Literal["labels"] = "labels"


class RelationshipsDataSource(StandardDataSource):
    _destination_type = "relationships"
    type: Literal["relationships"] = "relationships"


class DataSetsDataSource(StandardDataSource):
    _destination_type = "data_sets"
    type: Literal["data_sets"] = "data_sets"


class DataModelSource(Destination):
    _destination_type = "instances"
    type: Literal["instances"] = "instances"
    data_model: DataModelInfo = Field(description="Target data model info.")
    instance_space: str | None = Field(None, description="The space where the instances will be created.")


class ViewDataSource(Destination):
    _destination_type: ClassVar[str]
    type: Literal["nodes", "edges"]
    view: ViewInfo | None = Field(default=None, description="Target view info.")
    edge_type: EdgeType | None = Field(default=None, description="Target type of the connection definition.")
    instance_space: str | None = Field(
        default=None, description="The space where the instances(nodes/edges) will be created."
    )


class NodeViewDataSource(ViewDataSource):
    _destination_type = "nodes"
    type: Literal["nodes"] = "nodes"


class EdgeViewDataSource(ViewDataSource):
    _destination_type = "edges"
    type: Literal["edges"] = "edges"


class RawDataSource(Destination):
    _destination_type = "raw"
    type: Literal["raw"] = "raw"
    database: str = Field(description="The database name.")
    table: str = Field(description="The table name.")


class SequenceRowDataSource(Destination):
    _destination_type = "sequence_rows"
    type: Literal["sequence_rows"] = "sequence_rows"
    external_id: str = Field(description="The externalId of sequence.")


def get_all_destination_leaf_classes(base_class: type[Destination]) -> list:
    subclasses = base_class.__subclasses__()
    result = []

    if not subclasses:
        if base_class is not Destination:
            result.append(base_class)
    else:
        for subclass in subclasses:
            result.extend(get_all_destination_leaf_classes(subclass))

    return result


_DESTINATION_CLASS_BY_TYPE: MappingProxyType[str, type[Destination]] = MappingProxyType(
    {
        cls._destination_type: cls
        for cls in get_all_destination_leaf_classes(Destination)
        if hasattr(cls, "_destination_type") and cls._destination_type is not None
    }
)

ALL_DESTINATION_TYPES = sorted(_DESTINATION_CLASS_BY_TYPE)



import os
from collections.abc import Iterable, Mapping
from dataclasses import Field, dataclass, field, fields
from typing import Any, Literal, TypeAlias, get_args

import questionary
import typer
from cognite.client.credentials import (
    CredentialProvider,
    OAuthClientCredentials,
    OAuthDeviceCode,
    OAuthInteractive,
    Token,
)
from questionary import Choice
from rich import print

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.constants import TOOLKIT_CLIENT_ENTRA_ID
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitKeyError, ToolkitMissingValueError
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._version import __version__

Provider: TypeAlias = Literal["entra_id", "auth0", "cdf", "other"]
LoginFlow: TypeAlias = Literal["client_credentials", "token", "device_code", "interactive"]
VALID_PROVIDERS = get_args(Provider)
VALID_LOGIN_FLOWS = get_args(LoginFlow)

CLIENT_NAME = f"CDF-Toolkit:{__version__}"
PROVIDER_DESCRIPTION = {
    "entra_id": "Use Microsoft Entra ID to authenticate",
    "auth0": "Use Auth0 to authenticate",
    "cdf": "Use Cognite IDP to authenticate",
    "other": "Use other IDP to authenticate",
}
LOGIN_FLOW_DESCRIPTION = {
    "client_credentials": "Setup a service principal with client credentials",
    "interactive": "Login using the browser with your user credentials",
    "device_code": "Login using the browser with your user credentials using device code flow",
    "token": "Use a Token directly to authenticate",
}


@dataclass
class EnvOptions(Mapping):
    display_name: str
    default_example: str = ""
    example: dict[Provider, str] = field(default_factory=dict)
    is_secret: bool = False
    required: frozenset[tuple[Provider | None, LoginFlow]] = frozenset()
    optional: frozenset[tuple[Provider | None, LoginFlow]] = frozenset()

    def __getitem__(self, key: str) -> str | bool:
        return self.__dict__[key]

    def __iter__(self) -> Iterable[str]:  # type: ignore[override]
        return iter(self.__dict__.keys())

    def __len__(self) -> int:
        return len(self.__dict__)


ALL_CASES = [(None, flow) for flow in VALID_LOGIN_FLOWS]


def all_providers(
    flow: LoginFlow, exclude: Provider | Iterable[Provider] | None = None
) -> frozenset[tuple[Provider | None, LoginFlow]]:
    if isinstance(exclude, str):
        exclude = {exclude}
    elif exclude is not None:
        exclude = set(exclude)
    return frozenset((prov, flow) for prov in VALID_PROVIDERS if exclude is None or prov not in exclude)


@dataclass
class EnvironmentVariables:
    CDF_CLUSTER: str = field(metadata=EnvOptions("CDF cluster", "westeurope-1"))
    CDF_PROJECT: str = field(metadata=EnvOptions("CDF project", "publicdata"))
    PROVIDER: Provider = field(default="entra_id", metadata=EnvOptions("Provider", "entra_id"))
    LOGIN_FLOW: LoginFlow = field(default="client_credentials", metadata=EnvOptions("Login flow", "client_credentials"))
    CDF_URL: str | None = field(
        default=None,
        metadata=EnvOptions(
            "CDF URL", default_example="https://{CDF_CLUSTER}.cognitedata.com", optional=frozenset(ALL_CASES)
        ),
    )
    CDF_TOKEN: str | None = field(
        default=None, metadata=EnvOptions("OAuth2 token", required=frozenset([(None, "token")]))
    )
    IDP_CLIENT_ID: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="client id",
            required=frozenset(
                [(None, "client_credentials"), (None, "interactive"), *all_providers("device_code", exclude="entra_id")]
            ),
        ),
    )
    IDP_CLIENT_SECRET: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="client secret", is_secret=True, required=frozenset({(None, "client_credentials")})
        ),
    )
    IDP_TOKEN_URL: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="token URL",
            example={
                "entra_id": "https://login.microsoftonline.com/{IDP_TENANT_ID}/oauth2/v2.0/token",
                "auth0": "https://<my_auth_url>/oauth/token",
            },
            required=all_providers(flow="client_credentials", exclude={"entra_id", "cdf"}),
            optional=frozenset([("entra_id", "client_credentials")]),
        ),
    )
    IDP_TENANT_ID: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="Tenant id for MS Entra",
            example={"entra_id": "00000000-0000-0000-0000-000000000000 or mytenant.onmicrosoft.com"},
            required=frozenset(
                [("entra_id", "device_code"), ("entra_id", "client_credentials"), ("entra_id", "interactive")]
            ),
        ),
    )
    IDP_AUDIENCE: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="IDP audience",
            example={
                "entra_id": "https://{CDF_CLUSTER}.cognitedata.com",
                "auth0": "https: //{CDF_PROJECT}.fusion.cognite.com/{CDF_PROJECT}",
                "other": "https://{CDF_CLUSTER}.cognitedata.com",
            },
            required=all_providers(flow="client_credentials", exclude={"entra_id", "auth0", "cdf"}),
            optional=frozenset([("entra_id", "client_credentials"), ("auth0", "client_credentials")]),
        ),
    )

    IDP_SCOPES: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="IDP scopes",
            example={
                "entra_id": "https://{CDF_CLUSTER}.cognitedata.com/.default",
                "auth0": "IDENTITY,user_impersonation",
            },
            optional=frozenset([*all_providers("client_credentials", exclude="cdf"), (None, "interactive")]),
        ),
    )
    IDP_AUTHORITY_URL: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="IDP authority URL",
            example={"entra_id": "https://login.microsoftonline.com/{IDP_TENANT_ID}"},
            required=all_providers("interactive", exclude="entra_id"),
            optional=frozenset([("entra_id", "interactive"), ("auth0", "device_code")]),
        ),
    )
    IDP_DISCOVERY_URL: str | None = field(
        default=None,
        metadata=EnvOptions(
            display_name="IDP OIDC discovery URL (root URL excl. /.well-known/...)",
            default_example="https://<auth0-tenant>.auth0.com/oauth",
            required=all_providers("device_code", exclude="entra_id"),
        ),
    )
    CDF_CLIENT_TIMEOUT: int = field(
        default=30,
        metadata=EnvOptions(display_name="CDF client timeout", default_example="30", optional=frozenset(ALL_CASES)),
    )
    CDF_CLIENT_MAX_WORKERS: int = field(
        default=5,
        metadata=EnvOptions(display_name="CDF client max workers", default_example="5", optional=frozenset(ALL_CASES)),
    )
    _client: ToolkitClient | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        self.LOGIN_FLOW = self.LOGIN_FLOW.lower()  # type: ignore[assignment]
        if self.LOGIN_FLOW not in VALID_LOGIN_FLOWS:
            raise AuthenticationError(f"Invalid login flow: {self.LOGIN_FLOW}. Valid options are {VALID_LOGIN_FLOWS}")
        if self.PROVIDER not in VALID_PROVIDERS:
            raise AuthenticationError(f"Invalid provider: {self.PROVIDER}. Valid options are {VALID_PROVIDERS}")

    # All derived properties
    @property
    def idp_tenant_id(self) -> str:
        if self.IDP_TENANT_ID:
            return self.IDP_TENANT_ID
        if self.PROVIDER == "entra_id" and self.IDP_TOKEN_URL:
            return self.IDP_TOKEN_URL.removeprefix("https://login.microsoftonline.com/").removesuffix(
                "/oauth2/v2.0/token"
            )
        raise ToolkitMissingValueError("IDP_TENANT_ID is missing", "IDP_TENANT_ID")

    @property
    def idp_token_url(self) -> str:
        if self.PROVIDER == "cdf":
            return "https://auth.cognite.com/oauth2/token"
        if self.IDP_TOKEN_URL:
            return self.IDP_TOKEN_URL
        if self.PROVIDER == "entra_id" and self.IDP_TENANT_ID:
            return f"https://login.microsoftonline.com/{self.IDP_TENANT_ID}/oauth2/v2.0/token"
        alternative = ""
        if self.PROVIDER == "entra_id":
            alternative = " or provide IDP_TENANT_ID"
        raise ToolkitMissingValueError(
            f"IDP_TOKEN_URL is missing. Please provide it{alternative} in the environment variables.",
            "IDP_TOKEN_URL",
        )

    @property
    def cdf_url(self) -> str:
        return self.CDF_URL or f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_audience(self) -> str:
        if self.IDP_AUDIENCE:
            return self.IDP_AUDIENCE
        if self.PROVIDER == "auth0":
            return f"https://{self.CDF_PROJECT}.fusion.cognite.com/{self.CDF_PROJECT}"
        else:
            return f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_scopes(self) -> list[str]:
        if self.IDP_SCOPES:
            return self.IDP_SCOPES.split(",")
        if self.PROVIDER == "auth0":
            return ["IDENTITY", "user_impersonation"]
        return [f"https://{self.CDF_CLUSTER}.cognitedata.com/.default"]

    @property
    def idp_authority_url(self) -> str:
        if self.IDP_AUTHORITY_URL:
            return self.IDP_AUTHORITY_URL
        if self.PROVIDER == "entra_id" and self.idp_tenant_id:
            return f"https://login.microsoftonline.com/{self.idp_tenant_id}"
        alternative = ""
        if self.PROVIDER == "entra_id":
            alternative = " or provide IDP_TENANT_ID"
        raise ToolkitMissingValueError(
            f"IDP_AUTHORITY_URL is missing. Please provide it{alternative} in the environment variables.",
            "IDP_AUTHORITY_URL",
        )

    @classmethod
    def _fields(cls, inst: "EnvironmentVariables | None" = None) -> tuple[Field, ...]:
        return tuple(f for f in fields(inst or cls) if not f.name.startswith("_"))

    @classmethod
    def create_from_environment(cls) -> "EnvironmentVariables":
        if missing := [key for key in ["CDF_CLUSTER", "CDF_PROJECT"] if key not in os.environ]:
            raise ToolkitMissingValueError(f"Missing environment variables: {humanize_collection(missing)}")
        args: dict[str, Any] = {
            field_.name: field_.type(os.environ[field_.name]) if field_.type is int else os.environ[field_.name]  # type: ignore[operator]
            for field_ in cls._fields()
            if field_.name in os.environ
        }
        return cls(**args)

    def get_credentials(self) -> CredentialProvider:
        method_by_flow = {
            "client_credentials": self._get_oauth_client_credentials,
            "interactive": self._get_oauth_interactive,
            "device_code": self._get_oauth_device_code,
            "token": self._get_token,
        }
        if self.LOGIN_FLOW not in method_by_flow:
            # Should already be checked in __post_init__
            raise AuthenticationError(f"Login flow {self.LOGIN_FLOW} is not supported.")

        if missing_vars := self.get_missing_vars():
            raise ToolkitMissingValueError(
                f"The login flow '{self.LOGIN_FLOW}' requires the following environment variables: {humanize_collection(missing_vars)}.",
            )
        return method_by_flow[self.LOGIN_FLOW]()

    def _get_oauth_client_credentials(self) -> OAuthClientCredentials:
        if self.PROVIDER == "cdf":
            return OAuthClientCredentials(
                client_id=self.IDP_CLIENT_ID,  # type: ignore[arg-type]
                client_secret=self.IDP_CLIENT_SECRET,  # type: ignore[arg-type]
                token_url=self.idp_token_url,
                scopes=None,  # type: ignore[arg-type]
            )
        return OAuthClientCredentials(
            client_id=self.IDP_CLIENT_ID,  # type: ignore[arg-type]
            client_secret=self.IDP_CLIENT_SECRET,  # type: ignore[arg-type]
            token_url=self.idp_token_url,
            audience=self.idp_audience,
            scopes=self.idp_scopes,
        )

    def _get_oauth_interactive(self) -> OAuthInteractive:
        return OAuthInteractive(
            client_id=self.IDP_CLIENT_ID,  # type: ignore[arg-type]
            authority_url=self.idp_authority_url,
            scopes=self.idp_scopes,
        )

    def _get_oauth_device_code(self) -> OAuthDeviceCode:
        if self.PROVIDER == "entra_id":
            # TODO: If the user has submitted the wrong scopes, we may get a valid token that gives 401 on the CDF API.
            # The user will then have to wait until the token has expired to retry with the correct scopes.
            # If we add clear_cache=True to the OAuthDeviceCode, the token cache will be cleared.
            # We could add a cli option to auth verify, e.g. --clear-token-cache, that will clear the cache.
            return OAuthDeviceCode.default_for_azure_ad(
                tenant_id=self.IDP_TENANT_ID,  # type: ignore[arg-type]
                client_id=TOOLKIT_CLIENT_ENTRA_ID,
                cdf_cluster=self.CDF_CLUSTER,
                clear_cache=False,
            )
        return OAuthDeviceCode(
            authority_url=self.IDP_AUTHORITY_URL,
            cdf_cluster=self.CDF_CLUSTER,
            oauth_discovery_url=self.IDP_DISCOVERY_URL,
            client_id=self.IDP_CLIENT_ID,  # type: ignore[arg-type]
            audience=self.idp_audience,
        )

    def _get_token(self) -> Token:
        if not self.CDF_TOKEN:
            raise ToolkitKeyError("CDF_TOKEN must be set in the environment", "CDF_TOKEN")
        return Token(self.CDF_TOKEN)

    def get_config(self, is_strict_validation: bool) -> ToolkitClientConfig:
        return ToolkitClientConfig(
            client_name=CLIENT_NAME,
            project=self.CDF_PROJECT,
            credentials=self.get_credentials(),
            base_url=self.cdf_url,
            is_strict_validation=is_strict_validation,
            timeout=self.CDF_CLIENT_TIMEOUT,
            max_workers=self.CDF_CLIENT_MAX_WORKERS,
        )

    def get_client(self, is_strict_validation: bool = True, enable_set_pending_ids: bool = False) -> ToolkitClient:
        if self._client is None:
            self._client = ToolkitClient(
                config=self.get_config(is_strict_validation), enable_set_pending_ids=enable_set_pending_ids
            )
        return self._client

    def dump(self, include_os: bool = True) -> dict[str, str | None]:
        variables: dict[str, Any] = {}
        if include_os:
            variables.update(os.environ)
        for field_ in self._fields(self):
            value = self._get_value(field_)
            if isinstance(value, list):
                value = ",".join(value)
            if value is not None:
                if field_.type is int:
                    variables[field_.name] = value
                else:
                    variables[field_.name] = str(value)
        return variables

    def as_string(self) -> str:
        env_lines: list[str] = [f"CDF_URL={self.cdf_url}"]
        body = "\n".join(env_lines)
        return f"CDF Project {self.CDF_PROJECT!r} in cluster {self.CDF_CLUSTER!r}:\n{body}"

    def get_missing_vars(self) -> set[str]:
        provider, flow = self.PROVIDER, self.LOGIN_FLOW
        missing: set[str] = set()
        for field_ in self._fields(self):
            required = field_.metadata["required"]
            value = getattr(self, field_.name)
            if value is None and required and ((provider, flow) in required or (None, flow) in required):
                missing.add(field_.name)

        # Special cases, if IDP_TENANT_ID is missing.
        if (provider, flow) == ("entra_id", "client_credentials") and "IDP_TENANT_ID" in missing and self.IDP_TOKEN_URL:
            missing -= {"IDP_TENANT_ID"}
        if (provider, flow) == ("entra_id", "interactive") and "IDP_TENANT_ID" in missing and self.IDP_AUTHORITY_URL:
            missing -= {"IDP_TENANT_ID"}
        return missing

    def get_required_with_value(self, lookup_default: bool = False) -> list[tuple[Field, Any]]:
        provider, flow = self.PROVIDER, self.LOGIN_FLOW
        values: list[tuple[Field, Any]] = []
        for field_ in self._fields(self):
            required = field_.metadata["required"]
            if required and ((provider, flow) in required or (None, flow) in required):
                if field_.name == "IDP_TOKEN_URL" and provider == "entra_id":
                    continue
                value = self._get_value(field_, lookup_default)
                values.append((field_, value))
        return values

    def _get_value(self, field_: Field, lookup_default: bool = True) -> Any:
        if lookup_default and (default_name := field_.name.casefold()):
            try:
                if hasattr(self, default_name):
                    return getattr(self, default_name)
            except ToolkitMissingValueError:
                ...
        return getattr(self, field_.name)

    def get_optional_with_value(self) -> list[tuple[Field, Any]]:
        provider, flow = self.PROVIDER, self.LOGIN_FLOW
        values: list[tuple[Field, Any]] = []
        for field_ in self._fields(self):
            optional = field_.metadata["optional"]
            if optional and ((provider, flow) in optional or (None, flow) in optional):
                value = self._get_value(field_)
                values.append((field_, value))
        return values

    def create_dotenv_file(self) -> str:
        lines = [
            "# .env file generated by cognite-toolkit",
            f"CDF_CLUSTER={self.CDF_CLUSTER}",
            f"CDF_PROJECT={self.CDF_PROJECT}",
        ]
        if self.LOGIN_FLOW != "token":
            lines += [
                f"PROVIDER={self.PROVIDER}",
            ]
        lines += [
            f"LOGIN_FLOW={self.LOGIN_FLOW}",
        ]
        lines.append("")
        lines.append("# Required variables")
        for field_, value in self.get_required_with_value(lookup_default=True):
            if value is not None:
                lines.append(f"{field_.name}={value}")
        lines.append("")
        lines.append("# Optional variables (derived from the required variables)")
        for field_, value in self.get_optional_with_value():
            if value is None:
                continue
            if isinstance(value, list):
                value = ",".join(value)
            lines.append(f"{field_.name}={value}")
        return "\n".join(lines) + "\n"


def prompt_user_environment_variables(current: EnvironmentVariables | None = None) -> EnvironmentVariables:
    provider = questionary.select(
        "Choose the provider (Who authenticates you?)",
        choices=[
            Choice(title=f"{provider}: {description}", value=provider)
            # We CDF as a provider is not yet out of alpha, so we hide it from the user.
            for provider, description in PROVIDER_DESCRIPTION.items()
            if provider != "cdf"
        ],
        default=current.PROVIDER if current else "entra_id",
    ).ask()
    exclude = set()
    if provider == "cdf":
        exclude = set(VALID_LOGIN_FLOWS) - {"client_credentials"}
    choices = [
        Choice(title=f"{flow}: {description}", value=flow)
        for flow, description in LOGIN_FLOW_DESCRIPTION.items()
        if flow not in exclude
    ]
    if len(choices) == 1:
        print(f"Only one login flow available: {choices[0].title}")
        login_flow = choices[0].value
    else:
        login_flow = questionary.select(
            "Choose the login flow (How do you going to authenticate?)",
            choices=choices,
            default=current.LOGIN_FLOW if current else "client_credentials",
        ).ask()

    cdf_cluster = questionary.text("Enter the CDF cluster", default=current.CDF_CLUSTER if current else "").ask()
    cdf_project = questionary.text("Enter the CDF project", default=current.CDF_PROJECT if current else "").ask()
    args: dict[str, Any] = (
        current.dump(include_os=False)
        if current and _is_unchanged(current, provider, login_flow, cdf_project, cdf_cluster)  # type: ignore[arg-type]
        else {}
    )
    args.update(
        {"LOGIN_FLOW": login_flow, "CDF_CLUSTER": cdf_cluster, "CDF_PROJECT": cdf_project, "PROVIDER": provider}
    )
    env_vars = EnvironmentVariables(**args)
    idp_tenant_id = env_vars.IDP_TENANT_ID or "IDP_TENANT_ID"
    for field_, value in env_vars.get_required_with_value():
        user_value = get_user_value(field_, value, provider, cdf_cluster, cdf_project, idp_tenant_id)
        setattr(env_vars, field_.name, user_value)
        if field_.name == "IDP_TENANT_ID":
            idp_tenant_id = user_value

    optional_values = env_vars.get_optional_with_value()
    for field_, value in optional_values:
        print(f"  {field_.name}={value}")
    if questionary.confirm("Do you want to change any of these variables?", default=False).ask():
        for field_, value in optional_values:
            user_value = get_user_value(field_, value, provider, cdf_cluster, cdf_project, idp_tenant_id)
            setattr(env_vars, field_.name, user_value)
    return env_vars


def get_user_value(
    field_: Field, value: Any, provider: Provider, cdf_cluster: str, cdf_project: str, idp_tenant_id: str
) -> Any:
    is_secret = field_.metadata["is_secret"]
    display_name = field_.metadata["display_name"]
    if value:
        default = value
    else:
        default_example = field_.metadata["default_example"]
        default = (
            field_.metadata["example"]
            .get(provider, default_example)
            .format(CDF_CLUSTER=cdf_cluster, CDF_PROJECT=cdf_project, IDP_TENANT_ID=idp_tenant_id)
        )

    if isinstance(value, list):
        default = ",".join(value)
    elif value is not None and not isinstance(value, str):
        default = str(value)
    if is_secret:
        user_value = questionary.password(f"Enter the {display_name}:", default=default).ask()
    else:
        user_value = questionary.text(f"Enter the {display_name}:", default=default).ask()
    if user_value is None:
        raise typer.Exit(0)
    if field_.type is int:
        try:
            user_value = int(user_value)
        except ValueError:
            print(f"Invalid value: {user_value}. Please enter an integer.")
            return get_user_value(field_, value, provider, cdf_cluster, cdf_project, idp_tenant_id)
    return user_value


def _is_unchanged(
    current: EnvironmentVariables, provider: Provider, login_flow: LoginFlow, cdf_project: str, cdf_cluster: str
) -> bool:
    return (
        current.PROVIDER == provider
        and current.LOGIN_FLOW == login_flow
        and current.CDF_PROJECT == cdf_project
        and current.CDF_CLUSTER == cdf_cluster
    )


if __name__ == "__main__":
    # For easy testing
    envs = prompt_user_environment_variables()
    print(envs)



from collections.abc import Callable
from pathlib import Path

from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning

from ._base import Builder, DefaultBuilder, get_loader
from ._datamodels import DataModelBuilder
from ._file import FileBuilder
from ._function import FunctionBuilder
from ._location import LocationBuilder
from ._raw import RawBuilder
from ._streamlit import StreamlitBuilder
from ._transformation import TransformationBuilder


def create_builder(
    resource_folder: str,
    build_dir: Path | None,
    warn: Callable[[ToolkitWarning], None] | None = None,
) -> Builder:
    if builder_cls := _BUILDER_BY_RESOURCE_FOLDER.get(resource_folder):
        return builder_cls(build_dir, warn=warn)  # type: ignore[abstract]

    return DefaultBuilder(build_dir, resource_folder, warn)


_BUILDER_BY_RESOURCE_FOLDER = {_builder._resource_folder: _builder for _builder in Builder.__subclasses__()}
__all__ = [
    "Builder",
    "DataModelBuilder",
    "DefaultBuilder",
    "FileBuilder",
    "FunctionBuilder",
    "LocationBuilder",
    "RawBuilder",
    "StreamlitBuilder",
    "TransformationBuilder",
    "create_builder",
    "get_loader",
]



"""This is the core functionality of the Cognite Data Fusion Toolkit."""

import contextlib
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Optional, Union

import typer
from dotenv import load_dotenv
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import BuildCommand, CleanCommand, DeployCommand
from cognite_toolkit._cdf_tk.commands.clean import AVAILABLE_DATA_TYPES
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.utils import get_cicd_environment
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._version import __version__ as current_version


# Common parameters handled in common callback
@dataclass
class Common:
    override_env: bool


CDF_TOML = CDFToml.load(Path.cwd())


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"CDF-Toolkit version: {current_version}.")
        raise typer.Exit()


class CoreApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.common)
        self.command()(self.build)
        self.command()(self.deploy)
        self.command()(self.clean)

    def common(
        self,
        ctx: typer.Context,
        override_env: Annotated[
            bool,
            typer.Option(
                help="Load the .env file in this or the parent directory, but also override currently set environment variables",
            ),
        ] = True,
        env_path: Annotated[
            Optional[str],
            typer.Option(
                help="Path to .env file to load. Defaults to .env in current or parent directory.",
            ),
        ] = None,
        version: Annotated[
            bool,
            typer.Option(
                "--version",
                help="See which version of the Toolkit is installed.",
                callback=_version_callback,
            ),
        ] = False,
    ) -> None:
        """
        Docs: https://docs.cognite.com/cdf/deploy/cdf_toolkit/\n
        Template reference documentation: https://developer.cognite.com/sdks/toolkit/references/configs
        """
        if ctx.invoked_subcommand is None:
            print(
                Panel(
                    "\n".join(
                        [
                            "The Cognite Data Fusion Toolkit supports configuration of CDF projects from the command line or in CI/CD pipelines.",
                            "",
                            "[bold]Setup:[/]",
                            "1. Run [underline]cdf repo init[/] [italic]<directory name>[/] to set up a work directory.",
                            "2. Run [underline]cdf modules init[/] [italic]<directory name>[/] to initialise configuration modules.",
                            "",
                            "[bold]Configuration steps:[/]",
                            "3. Run [underline]cdf build[/] [italic]<directory name>[/] to verify the configuration for your project. Repeat for as many times as needed.",
                            "   Tip:[underline]cdf modules list[/] [italic]<directory name>[/] gives an overview of all your modules and their status.",
                            "",
                            "[bold]Deployment steps:[/]",
                            "4. Commit the [italic]<directory name>[/] to version control",
                            "5. Run [underline]cdf auth verify --interactive[/] to check that you have access to the relevant CDF project. ",
                            "    or [underline]cdf auth verify[/] if you have a .env file",
                            "6. Run [underline]cdf deploy --dry-run[/] to simulate the deployment of the configuration to the CDF project. Review the report provided.",
                            "7. Run [underline]cdf deploy[/] to deploy the configuration to the CDF project.",
                        ]
                    ),
                    title="Getting started",
                    style="green",
                    padding=(1, 2),
                )
            )
            return
        if override_env:
            print("  [bold yellow]WARNING:[/] Overriding environment variables with values from .env file...")

        if env_path is not None:
            if not (dotenv_file := Path(env_path)).is_file():
                raise ToolkitFileNotFoundError(env_path)

        else:
            if not (dotenv_file := Path.cwd() / ".env").is_file():
                if not (dotenv_file := Path.cwd().parent / ".env").is_file():
                    if get_cicd_environment() == "local":
                        warn = False
                        try:
                            env_vars = EnvironmentVariables.create_from_environment()
                        except Exception:
                            warn = True
                        else:
                            warn = bool(env_vars.get_missing_vars())
                        if warn:
                            print("[bold yellow]WARNING:[/] No .env file found in current or parent directory.")

        if dotenv_file.is_file():
            has_loaded = load_dotenv(dotenv_file, override=override_env)
            if not has_loaded:
                print("  [bold yellow]WARNING:[/] No environment variables found in .env file.")

        ctx.obj = Common(override_env=override_env)

    def build(
        self,
        ctx: typer.Context,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Where to find the module templates to build from",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        build_dir: Annotated[
            Path,
            typer.Option(
                "--build-dir",
                "-b",
                help="Where to save the built module files",
            ),
        ] = Path("./build"),
        selected: Annotated[
            Optional[list[str]],
            typer.Option(
                "--modules",
                "-m",
                help="Specify paths or names to the modules to build",
            ),
        ] = None,
        build_env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="The name of the environment to build",
            ),
        ] = CDF_TOML.cdf.default_env,
        no_clean: Annotated[
            bool,
            typer.Option(
                "--no-clean",
                "-c",
                help="Whether not to delete the build directory before building the configurations",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
        offline: Annotated[
            bool,
            typer.Option(
                "--offline",
                help="Do not check CDF for missing dependencies.",
            ),
        ] = False,
        exit_on_warning: Annotated[
            bool,
            typer.Option(
                "--exit-non-zero-on-warning",
                "-w",
                help="Exit with non-zero code on warning.",
                hidden=not Flags.EXIT_ON_WARNING.is_enabled(),
            ),
        ] = False,
    ) -> None:
        """Build configuration files from the modules to the build directory."""
        client: Union[ToolkitClient, None] = None
        if not offline:
            with contextlib.redirect_stdout(None), contextlib.suppress(Exception):
                # Remove the Error message from failing to load the config
                # This is verified in check_auth
                client = EnvironmentVariables.create_from_environment().get_client()

        print_warning = True
        if Flags.EXIT_ON_WARNING.is_enabled() and exit_on_warning:
            print_warning = False

        cmd = BuildCommand(print_warning=print_warning)
        cmd.run(
            lambda: cmd.execute(
                verbose,
                organization_dir,
                build_dir,
                selected,  # type: ignore[arg-type]
                build_env_name,
                no_clean,
                client,
                on_error="raise",
            )
        )

        if Flags.EXIT_ON_WARNING.is_enabled() and exit_on_warning and cmd.warning_list:
            print("\n[bold red]Warnings raised during the build process:[/]\n")

            for warning in cmd.warning_list:
                warning.print_warning(include_timestamp=False)
                print(end="\n")

            raise typer.Exit(code=1)

    def deploy(
        self,
        ctx: typer.Context,
        build_dir: Annotated[
            Path,
            typer.Argument(
                help="Where to find the module templates to deploy from. Defaults to current directory.",
                allow_dash=True,
            ),
        ] = Path("./build"),
        build_env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="CDF project environment to use for deployment. This is optional and "
                "if passed it is used to verify against the build environment",
            ),
        ] = None,
        drop: Annotated[
            bool,
            typer.Option(
                "--drop",
                "-d",
                help="Whether to drop existing configurations, drop per resource if present.",
            ),
        ] = False,
        drop_data: Annotated[
            bool,
            typer.Option(
                "--drop-data",
                help="Whether to drop existing data in data model containers and spaces.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        include: Annotated[
            Optional[list[str]],
            typer.Option(
                "--include",
                help=f"Specify which resources to deploy, available options: {AVAILABLE_DATA_TYPES}.",
            ),
        ] = None,
        force_update: Annotated[
            bool,
            typer.Option(
                "--force-update",
                help="Whether to force update the resources in the CDF project even if they are considered unchanged.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Deploys the configuration files in the build directory to the CDF project."""
        cmd = DeployCommand(print_warning=True)
        env_vars = EnvironmentVariables.create_from_environment()
        cmd.run(
            lambda: cmd.deploy_build_directory(
                env_vars=env_vars,
                build_dir=build_dir,
                build_env_name=build_env_name,
                dry_run=dry_run,
                drop=drop,
                drop_data=drop_data,
                force_update=force_update,
                include=include,
                verbose=verbose,
            )
        )

    def clean(
        self,
        ctx: typer.Context,
        build_dir: Annotated[
            Path,
            typer.Argument(
                help="Where to find the module templates to clean from. Defaults to ./build directory.",
                allow_dash=True,
            ),
        ] = Path("./build"),
        build_env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="CDF project environment to use for cleaning. This is optional and "
                "if passed it is used to verify against the build environment",
            ),
        ] = None,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present",
            ),
        ] = False,
        include: Annotated[
            Optional[list[str]],
            typer.Option(
                "--include",
                help=f"Specify which resource types to deploy, supported types: {AVAILABLE_DATA_TYPES}",
            ),
        ] = None,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Cleans the resources in the build directory from the CDF project."""
        # Override cluster and project from the options/env variables
        cmd = CleanCommand(print_warning=True)
        env = EnvironmentVariables.create_from_environment()
        cmd.run(
            lambda: cmd.execute(
                env,
                build_dir,
                build_env_name,
                dry_run,
                include,
                verbose,
            )
        )



from rich import print

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.deploy import DeployCommand
from cognite_toolkit._cdf_tk.data_classes import DeployResults
from cognite_toolkit._cdf_tk.loaders import (
    ContainerLoader,
    DataModelLoader,
    ResourceWorker,
    SpaceLoader,
    ViewLoader,
)

from .data_model import COGNITE_MIGRATION_MODEL, MAPPING_CONTAINER, MAPPING_VIEW, MODEL_ID, SPACE


class MigrationPrepareCommand(ToolkitCommand):
    def deploy_cognite_migration(self, client: ToolkitClient, dry_run: bool, verbose: bool = False) -> DeployResults:
        """Deploys the Cognite Migration Data Model"""

        deploy_cmd = DeployCommand(self.print_warning, silent=self.silent)
        deploy_cmd.tracker = self.tracker

        verb = "Would deploy" if dry_run else "Deploying"
        print(f"{verb} {MODEL_ID!r}")
        results = DeployResults([], "deploy", dry_run=dry_run)
        for loader_cls, resources in [
            (SpaceLoader, [SPACE]),
            (ContainerLoader, [MAPPING_CONTAINER]),
            (ViewLoader, [MAPPING_VIEW]),
            (DataModelLoader, [COGNITE_MIGRATION_MODEL]),
        ]:
            # MyPy does not understand that `loader_cls` has a `create_loader` method.
            loader = loader_cls.create_loader(client)  # type: ignore[attr-defined]
            worker = ResourceWorker(loader)
            # MyPy does not understand that `loader` has a `get_id` method.
            local_by_id = {loader.get_id(item): (item.dump(), item) for item in resources}  # type: ignore[attr-defined]
            worker.validate_access(local_by_id, is_dry_run=dry_run)
            cdf_resources = loader.retrieve(list(local_by_id.keys()))
            to_create, to_update, to_delete, unchanged = worker.categorize_resources(
                local_by_id, cdf_resources, False, verbose
            )

            if dry_run:
                result = deploy_cmd.dry_run_deploy(to_create, to_update, to_delete, unchanged, loader, False, False)
            else:
                result = deploy_cmd.actual_deploy(
                    to_create,
                    to_update,
                    to_delete,
                    unchanged,
                    loader,
                )
            if result:
                results[result.name] = result
        if results.has_counts:
            print(results.counts_table())
        return results



"""This module contains data classes for the APM Config node used to configure Infield.

These classes are used to represent the configuration of the APM Config node in a structured way, such that it can be
used in the InfieldV1 Loader and thus be represented as a resource type in Toolkit. We do not do any validation in the
FeatureConfiguration objects as this is just JSON object in the node, but use the structure to do lookup of
data sets, spaces, and groups.
"""

import sys
from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass, field, fields
from functools import lru_cache
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import Node, NodeApply, NodeApplyList, NodeOrEdgeData, ViewId
from cognite.client.utils._text import to_camel_case

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class APMConfigObject(CogniteObject):
    """This is used as a base class for all objects in the APM Config node.

    This ensures that all extra fields are  picked up when loading the resource from file and stored in the _extra field.
    When the request is created the dump method is called and the extra fields are added to the request.

    This is done to ensure that all fields are included in the request and that the request is valid in case JSON blob
    in the property featureConfiguration in the view (APM_Config, APM_Config, 1) is changed in the future.
    """

    _extra: dict[str, Any] = field(default_factory=dict, init=False)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        instance._extra = {k: v for k, v in resource.items() if k not in cls.get_field_names()}
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self._extra:
            output.update(self._extra)
        return output

    @classmethod
    @lru_cache(maxsize=1)
    def get_field_names(cls) -> set[str]:
        cls_fields = list(fields(cls))
        return {field.name for field in cls_fields} | {to_camel_case(field.name) for field in cls_fields}


@dataclass
class ThreeDModelIdentifier(APMConfigObject):
    revision_id: int | None = None
    model_id: int | None = None
    name: str | None = None


@dataclass
class ThreeDConfiguration(APMConfigObject):
    full_weight_models: list[ThreeDModelIdentifier] | None = None
    light_weight_models: list[ThreeDModelIdentifier] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        for snake, camel in [("full_weight_models", "fullWeightModels"), ("light_weight_models", "lightWeightModels")]:
            if camel in resource:
                setattr(
                    instance,
                    snake,
                    [ThreeDModelIdentifier._load(model, cognite_client=cognite_client) for model in resource[camel]],
                )
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        for snake, camel in [("full_weight_models", "fullWeightModels"), ("light_weight_models", "lightWeightModels")]:
            if hasattr(self, snake):
                output[camel if camel_case else snake] = [value.dump() for value in getattr(self, snake) or []]
        return output


@dataclass
class ResourceFilters(APMConfigObject):
    data_set_ids: list[int] | None = None
    asset_subtree_external_ids: list[str] | None = None
    root_asset_external_ids: list[str] | None = None
    external_id_prefix: str | None = None
    spaces: list[str] | None = None


@dataclass
class RootLocationDataFilters(APMConfigObject):
    general: ResourceFilters | None = None
    assets: ResourceFilters | None = None
    files: ResourceFilters | None = None
    timeseries: ResourceFilters | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        for key in ["general", "assets", "files", "timeseries"]:
            if key in resource:
                setattr(instance, key, ResourceFilters._load(resource[key], cognite_client=cognite_client))
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        for key in ["general", "assets", "files", "timeseries"]:
            if hasattr(self, key):
                if value := getattr(self, key):
                    output[key] = value.dump(camel_case=camel_case)
        return output


@dataclass
class ObservationFeatureToggles(APMConfigObject):
    is_enabled: bool | None = None
    is_write_back_enabled: bool | None = None
    notifications_endpoint_external_id: str | None = None
    attachments_endpoint_external_id: str | None = None


@dataclass
class RootLocationFeatureToggles(APMConfigObject):
    three_d: bool | None = None
    trends: bool | None = None
    documents: bool | None = None
    workorders: bool | None = None
    notifications: bool | None = None
    media: bool | None = None
    template_checklist_flow: bool | None = None
    workorder_checklist_flow: bool | None = None
    observations: ObservationFeatureToggles | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        if "observations" in resource:
            instance.observations = ObservationFeatureToggles._load(
                resource["observations"], cognite_client=cognite_client
            )
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.observations:
            output["observations"] = self.observations.dump(camel_case=camel_case)
        return output


@dataclass
class ObservationConfigFieldProperty(APMConfigObject):
    display_title: str | None = None
    display_description: str | None = None
    is_required: bool | None = None


@dataclass
class ObservationConfigDropdownPropertyOption(APMConfigObject):
    id: str | None = None
    value: str | None = None
    label: str | None = None


@dataclass
class ObservationConfigDropdownProperty(ObservationConfigFieldProperty):
    options: list[ObservationConfigDropdownPropertyOption] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        if "options" in resource:
            instance.options = [
                ObservationConfigDropdownPropertyOption._load(option, cognite_client=cognite_client)
                for option in resource["options"]
            ]
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.options:
            output["options"] = [option.dump(camel_case=camel_case) for option in self.options]
        return output


@dataclass
class ObservationsConfig(APMConfigObject):
    files: ObservationConfigFieldProperty | None = None
    description: ObservationConfigFieldProperty | None = None
    asset: ObservationConfigFieldProperty | None = None
    troubleshooting: ObservationConfigFieldProperty | None = None
    type: ObservationConfigDropdownProperty | None = None
    priority: ObservationConfigDropdownProperty | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        for key in ["files", "description", "asset", "troubleshooting"]:
            if key in resource:
                setattr(
                    instance, key, ObservationConfigFieldProperty._load(resource[key], cognite_client=cognite_client)
                )
        for key in ["type", "priority"]:
            if key in resource:
                setattr(
                    instance, key, ObservationConfigDropdownProperty._load(resource[key], cognite_client=cognite_client)
                )
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        for key in ["files", "description", "asset", "troubleshooting", "type", "priority"]:
            if hasattr(self, key):
                if value := getattr(self, key):
                    output[key] = value.dump(camel_case=camel_case)
        return output


@dataclass
class RootLocationConfiguration(APMConfigObject):
    asset_external_id: str | None = None
    external_id: str | None = None
    display_name: str | None = None
    three_d_configuration: ThreeDConfiguration | None = None
    data_set_id: int | None = None
    template_admins: list[str] | None = None  # list of Group Names
    checklist_admins: list[str] | None = None  # list of Group Names
    app_data_instance_space: str | None = None
    source_data_instance_space: str | None = None
    data_filters: RootLocationDataFilters | None = None
    feature_toggles: RootLocationFeatureToggles | None = None
    observations: ObservationsConfig | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        if "threeDConfiguration" in resource:
            instance.three_d_configuration = ThreeDConfiguration._load(
                resource["threeDConfiguration"], cognite_client=cognite_client
            )
        if "dataFilters" in resource:
            instance.data_filters = RootLocationDataFilters._load(
                resource["dataFilters"], cognite_client=cognite_client
            )
        if "featureToggles" in resource:
            instance.feature_toggles = RootLocationFeatureToggles._load(
                resource["featureToggles"], cognite_client=cognite_client
            )
        if "observations" in resource:
            instance.observations = ObservationsConfig._load(resource["observations"], cognite_client=cognite_client)
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.three_d_configuration:
            output["threeDConfiguration" if camel_case else "three_d_configuration"] = self.three_d_configuration.dump(
                camel_case=camel_case
            )
        if self.data_filters:
            output["dataFilters" if camel_case else "data_filters"] = self.data_filters.dump(camel_case=camel_case)
        if self.feature_toggles:
            output["featureToggles" if camel_case else "feature_toggles"] = self.feature_toggles.dump(
                camel_case=camel_case
            )
        if self.observations:
            output["observations"] = self.observations.dump(camel_case=camel_case)
        return output


@dataclass
class FeatureConfiguration(APMConfigObject):
    root_location_configurations: list[RootLocationConfiguration] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        instance = super()._load(resource, cognite_client=cognite_client)
        if "rootLocationConfigurations" in resource:
            instance.root_location_configurations = [
                RootLocationConfiguration._load(item, cognite_client=cognite_client)
                for item in resource.get("rootLocationConfigurations", [])
            ]
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.root_location_configurations:
            output["rootLocationConfigurations" if camel_case else "root_location_configurations"] = [
                item.dump(camel_case=camel_case) for item in self.root_location_configurations
            ]
        return output


class APMConfigCore(WriteableCogniteResource["APMConfigWrite"], ABC):
    space: str = "APM_Config"
    view_id: ViewId = ViewId("APM_Config", "APM_Config", "1")

    def __init__(
        self,
        external_id: str,
        name: str | None = None,
        app_data_space_id: str | None = None,
        app_data_space_version: str | None = None,
        customer_data_space_id: str | None = None,
        customer_data_space_version: str | None = None,
        feature_configuration: FeatureConfiguration | None = None,
    ) -> None:
        self.external_id = external_id
        self.name = name
        self.app_data_space_id = app_data_space_id
        self.app_data_space_version = app_data_space_version
        self.customer_data_space_id = customer_data_space_id
        self.customer_data_space_version = customer_data_space_version
        self.feature_configuration = feature_configuration

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.feature_configuration:
            output["featureConfiguration" if camel_case else "feature_configuration"] = self.feature_configuration.dump(
                camel_case=camel_case
            )
        return output


class APMConfigWrite(APMConfigCore):
    def __init__(
        self,
        external_id: str,
        name: str | None = None,
        app_data_space_id: str | None = None,
        app_data_space_version: str | None = None,
        customer_data_space_id: str | None = None,
        customer_data_space_version: str | None = None,
        feature_configuration: FeatureConfiguration | None = None,
        existing_version: int | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            app_data_space_id=app_data_space_id,
            app_data_space_version=app_data_space_version,
            customer_data_space_id=customer_data_space_id,
            customer_data_space_version=customer_data_space_version,
            feature_configuration=feature_configuration,
        )
        self.existing_version = existing_version

    def as_write(self) -> "APMConfigWrite":
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            name=resource.get("name"),
            app_data_space_id=resource.get("appDataSpaceId"),
            app_data_space_version=resource.get("appDataSpaceVersion"),
            customer_data_space_id=resource.get("customerDataSpaceId"),
            customer_data_space_version=resource.get("customerDataSpaceVersion"),
            feature_configuration=FeatureConfiguration._load(
                resource["featureConfiguration"], cognite_client=cognite_client
            )
            if "featureConfiguration" in resource
            else None,
            existing_version=resource.get("existingVersion"),
        )

    def as_node(self) -> NodeApply:
        properties = self.dump(camel_case=True)
        # Node properties.
        properties.pop("externalId", None)
        properties.pop("existingVersion", None)
        return NodeApply(
            space=self.space,
            external_id=self.external_id,
            sources=[
                NodeOrEdgeData(
                    source=self.view_id,
                    properties=properties,
                )
            ],
            existing_version=self.existing_version,
        )


class APMConfig(APMConfigCore):
    def __init__(
        self,
        external_id: str,
        version: int,
        created_time: int,
        last_updated_time: int,
        name: str | None = None,
        app_data_space_id: str | None = None,
        app_data_space_version: str | None = None,
        customer_data_space_id: str | None = None,
        customer_data_space_version: str | None = None,
        feature_configuration: FeatureConfiguration | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            app_data_space_id=app_data_space_id,
            app_data_space_version=app_data_space_version,
            customer_data_space_id=customer_data_space_id,
            customer_data_space_version=customer_data_space_version,
            feature_configuration=feature_configuration,
        )
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.version = version

    def as_write(self) -> APMConfigWrite:
        return APMConfigWrite(
            external_id=self.external_id,
            name=self.name,
            app_data_space_id=self.app_data_space_id,
            app_data_space_version=self.app_data_space_version,
            customer_data_space_id=self.customer_data_space_id,
            customer_data_space_version=self.customer_data_space_version,
            feature_configuration=self.feature_configuration,
            existing_version=self.version,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            version=resource["version"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            name=resource.get("name"),
            app_data_space_id=resource.get("appDataSpaceId"),
            app_data_space_version=resource.get("appDataSpaceVersion"),
            customer_data_space_id=resource.get("customerDataSpaceId"),
            customer_data_space_version=resource.get("customerDataSpaceVersion"),
            feature_configuration=FeatureConfiguration._load(
                resource["featureConfiguration"], cognite_client=cognite_client
            )
            if "featureConfiguration" in resource
            else None,
        )

    @classmethod
    def from_node(cls, node: Node) -> Self:
        if node.space != cls.space:
            raise ValueError(f"Wrong instance space: {node.space}. {cls.__name__} nodes must be in {cls.space} space.")
        view_identifier = {ViewId.load(identifier): identifier for identifier in node.properties.keys()}
        if cls.view_id not in view_identifier:
            raise ValueError(
                f"Missing {cls.__name__} properties. All {cls.__name__} nodes must have properties from {cls.view_id}."
            )
        identifier = view_identifier[cls.view_id]

        resource = dict(node.properties[identifier])
        resource["externalId"] = node.external_id
        resource["createdTime"] = node.created_time
        resource["lastUpdatedTime"] = node.last_updated_time
        resource["version"] = node.version
        return cls._load(resource, None)


class APMConfigWriteList(CogniteResourceList[APMConfigWrite]):
    _RESOURCE = APMConfigWrite

    def as_nodes(self) -> NodeApplyList:
        return NodeApplyList([item.as_node() for item in self])


class APMConfigList(WriteableCogniteResourceList[APMConfigWrite, APMConfig]):
    _RESOURCE = APMConfig

    def as_write(self) -> APMConfigWriteList:
        return APMConfigWriteList([item.as_write() for item in self])

    @classmethod
    def from_nodes(cls, node: Sequence[Node]) -> Self:
        return cls([cls._RESOURCE.from_node(item) for item in node], cognite_client=None)
